
import sys
from enum import IntEnum
from .. import dzn2dict

from queue import Queue


__all__ = ['Status', 'Solutions']


SOLN_SEP = '----------'
SEARCH_COMPLETE = '=========='
UNKNOWN = '=====UNKNOWN====='
UNSATISFIABLE = '=====UNSATISFIABLE====='
UNBOUNDED = '=====UNBOUNDED====='
UNSATorUNBOUNDED = '=====UNSATorUNBOUNDED====='
ERROR = '=====ERROR====='


class Status(IntEnum):
    """Status of the solution stream."""

    #: The solution stream is complete (all solutions for satisfaction problems,
    #: optimal solution for optimization problems)
    COMPLETE = 0

    #: The solution stream is incomplete
    INCOMPLETE = 1

    #: The solution stream is empty (no solution found in the time limit)
    UNKNOWN = 2

    #: The problem admits no solution
    UNSATISFIABLE = 3

    #: The problem admits infinite solutions (unbounded domain)
    UNBOUNDED = 4

    #: Either unsatisfiable or unbounded
    UNSATorUNBOUNDED = 5

    #: Generic error in the execution of the solver
    ERROR = 6


class Solutions:
    """Solution stream returned by the ``pymzn.minizinc`` function.

    You should not need to instantiate this class in any other way than by
    calling the ``pymzn.minizinc`` or the ``pymzn.aio.minizinc`` functions.
    This class represents lazy list-like objects that collect the solutions
    provided by the solver and parsed by the PyMzn solution parser. The solution
    parser provides solutions to this object through a queue that is only
    accessed when this object is addressed or iterated over. If the queue has
    limited size (by using the ``max_queue_size`` option of the
    ``pymzn.aio.minizinc`` function), the execution of the solver will halt
    untill this object is addressed. Note that, by default, as soon as this
    object is addressed, the *full* queue is processed and it is cached in
    memory. To avoid this behavior, use the option ``keep_solutions=True`` in
    the ``pymzn.minizinc`` or ``pymzn.aio.minizinc`` functions.

    Arguments
    ---------
    status : Status
        The status of the solution stream, i.e. whether it is complete, the
        problem was unsatisfiable or other errors that might have occurred.
    log : str
        The log of the solver on standard output. Usually contains solver
        statistics and other log messages.
    stderr : str
        The log of the MiniZinc executable on standard error. Usually contains
        log messages about the flattening process, statistics and error
        messages.
    """

    def __init__(self, queue, *, keep=True):
        self._queue = queue
        self._keep = keep
        self._solns = [] if keep else None
        self._n_solns = 0
        self.status = Status.INCOMPLETE
        self.log = None
        self.stderr = None

    def _fetch(self):
        while not self._queue.empty():
            soln = self._queue.get_nowait()
            if self._keep:
                self._solns.append(soln)
            self._n_solns += 1
            yield soln

    def _fetch_all(self):
        for soln in self._fetch():
            pass

    def __len__(self):
        if self._keep:
            self._fetch_all()
        return self._n_solns

    def __iter__(self):
        if self._keep:
            self._fetch_all()
            return iter(self._solns)
        else:
            return self._fetch()

    def __getitem__(self, key):
        if not self._keep:
            raise RuntimeError(
                'Cannot address directly if keep_solutions is False'
            )
        self._fetch_all()
        return self._solns[key]

    def _pp_solns(self):
        if len(self._solns) <= 1:
            return str(self._solns)
        pp = ['[']
        for i, soln in enumerate(self._solns):
            pp.append('    ' + repr(soln))
            if i < len(self._solns) - 1:
                pp[-1] += ','
        pp.append(']')
        return '\n'.join(pp)

    def __repr__(self):
        if self._keep and self.status < 2:
            self._fetch_all()
            if len(self._solns) > 0:
                return '<Solutions: {}>'.format(self._pp_solns())
        return '<Solutions: {}>'.format(self.status.name)

    def __str__(self):
        if self._keep and self.status < 2:
            self._fetch_all()
            if len(self._solns) > 0:
                return self._pp_solns()
        return self.status.name

    def print(self, output_file=sys.stdout, log=False):
        """Print the solution stream"""

        for soln in iter(self):
            print(soln, file=output_file)
            print(SOLN_SEP, file=output_file)

        if self.status == 0:
            print(SEARCH_COMPLETE, file=output_file)

        if (self.status == 1 and self._n_solns == 0) or self.status >= 2:
            print({
                Status.INCOMPLETE : ERROR,
                Status.UNKNOWN: UNKNOWN,
                Status.UNSATISFIABLE: UNSATISFIABLE,
                Status.UNBOUNDED: UNBOUNDED,
                Status.UNSATorUNBOUNDED: UNSATorUNBOUNDED,
                Status.ERROR: ERROR
            }[self.status], file=output_file)

            if self.stderr:
                print(self.stderr.strip(), file=sys.stderr)

        elif log:
            print(str(self.log), file=output_file)


class SolutionParser:

    def __init__(
        self, solver, output_mode='dict', rebase_arrays=True, types=None,
        keep_solutions=True, return_enums=False
    ):
        self.solver = solver
        self.solver_parser = self.solver.parser()
        self.output_mode = output_mode
        self.rebase_arrays = rebase_arrays
        self.types = types
        self.keep_solutions = keep_solutions
        self.return_enums = return_enums
        self.status = Status.INCOMPLETE

    def _collect(self, solns, proc):
        for soln in self._parse(proc):
            solns._queue.put(soln)
        solns.status = self.status
        solns.stderr = proc.stderr_data
        solns.log = self.solver_parser.log

    def parse(self, proc):
        solns = Solutions(Queue(), keep=self.keep_solutions)
        self._collect(solns, proc)
        return solns

    def _parse(self, proc):
        parse_lines = self._parse_lines()
        parse_lines.send(None)
        for line in proc.readlines():
            soln = parse_lines.send(line)
            if soln is not None:
                yield soln

    def _parse_lines(self):
        solver_parse_out = self.solver_parser.parse_out()
        split_solns = self._split_solns()
        solver_parse_out.send(None)
        split_solns.send(None)

        line = yield
        while True:
            line = solver_parse_out.send(line)
            soln = split_solns.send(line)
            if soln is not None:
                if self.output_mode == 'dict':
                    soln = dzn2dict(
                        soln, rebase_arrays=self.rebase_arrays,
                        types=self.types, return_enums=self.return_enums
                    )
                line = yield soln
            else:
                line = yield

    def _split_solns(self):
        _buffer = []
        line = yield
        while True:
            line = line.strip()
            if line == SOLN_SEP:
                line = yield '\n'.join(_buffer)
                _buffer = []
                continue
            elif line == SEARCH_COMPLETE:
                self.status = Status.COMPLETE
                _buffer = []
            elif line == UNKNOWN:
                self.status = Status.UNKNOWN
            elif line == UNSATISFIABLE:
                self.status = Status.UNSATISFIABLE
            elif line == UNBOUNDED:
                self.status = Status.UNBOUNDED
            elif line == UNSATorUNBOUNDED:
                self.status = Status.UNSATorUNBOUNDED
            elif line == ERROR:
                self.status = Status.ERROR
            elif line:
                _buffer.append(line)
            line = yield




# -*- coding: utf-8 -*-
"""\
PyMzn is a Python library that wraps and enhances the MiniZinc tools for
constraint programming. PyMzn is built on top of the minizinc toolkit and
provides a number of off-the-shelf functions to readily solve problems encoded
with the MiniZinc language and return solutions as Python dictionaries.
"""

from .log import *
from .config import config
from .dzn import *
from .mzn import *

__all__ = ['config'] + log.__all__ + dzn.__all__ + mzn.__all__


__version__ = '0.18.1'


def main():
    import sys
    import ast
    import argparse

    def _minizinc(
        no_declare_enums=False, solver=None, output_file=None, solver_log=False,
        **_args
    ):
        if no_declare_enums:
            _args['declare_enums'] = False

        if 'compile' in _args and _args['compile']:
            other_args = {
                k: v for k, v in _args.items() if k not in ['mzn', 'dzn_files']
            }
            mzn2fzn(_args['mzn'], *_args['dzn_files'], **other_args)
            return

        if solver:
            from .mzn import solvers
            _args['solver'] = getattr(solvers, solver)

        solns = minizinc(
            _args['mzn'], *_args['dzn_files'], keep_solutions=False,
            **{k: v for k, v in _args.items() if k not in ['mzn', 'dzn_files']}
        )

        if output_file:
            out = open(output_file, 'w+')
        else:
            out = sys.stdout

        if 'output_mode' in _args and _args['output_mode'] == 'raw':
            print(solns, file=out)
        else:
            solns.print(output_file=out, log=solver_log)
        out.close()

    def _config(key=None, value=None, delete=False, **__):
        if not key:
            print('\n'.join(list(config.keys())))
        elif delete:
            del config[key]
            config.dump()
        elif value is None:
            print('{} : "{}"'.format(key, config.get(key)))
        else:
            config[key] = value
            config.dump()

    desc = sys.modules[__name__].__doc__
    fmt = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(description=desc, formatter_class=fmt)
    parser.add_argument(
        '--version', action='version',
        version=(
            'PyMzn, version {} | Copyright (c) 2016 Paolo Dragone'
        ).format(__version__)
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='display informative messages'
    )

    subparsers = parser.add_subparsers()

    # minizinc
    minizinc_parser = subparsers.add_parser(
        'minizinc', help='execute minizinc'
    )
    minizinc_parser.add_argument(
        'mzn', help='the mzn file to solve'
    )

    flattener_options = minizinc_parser.add_argument_group('Flattener options')
    flattener_options.add_argument(
        'dzn_files', nargs='*', help='additional dzn files'
    )
    flattener_options.add_argument(
        '--data', type=ast.literal_eval, help='additional inline data'
    )
    flattener_options.add_argument(
        '--args', type=ast.literal_eval,
        help='additional arguments for the template engine'
    )
    flattener_options.add_argument(
        '--output-vars', type=ast.literal_eval,
        help='list of variables to output'
    )
    flattener_options.add_argument(
        '-I', '--include', dest='path', action='append',
        help='additional search directories'
    )
    flattener_options.add_argument(
        '--stdlib-dir', help='path to MiniZinc standard library directory'
    )
    flattener_options.add_argument(
        '-G', '--globals-dir', help='name of directory including globals'
    )
    flattener_options.add_argument(
        '-c', '--compile', action='store_true',
        help='compile only'
    )
    flattener_options.add_argument(
        '-k', '--keep', action='store_true',
        help='keep generated files'
    )
    flattener_options.add_argument(
        '--two-pass', type=int, help='equivalent to MiniZinc -O<n> option'
    )
    flattener_options.add_argument(
        '--pre-passes', type=int,
        help='equivalent to MiniZinc --pre-passes option'
    )

    solver_options = minizinc_parser.add_argument_group('Solver options')
    solver_options.add_argument(
        '-S', '--solver', help='name of the solver'
    )
    solver_options.add_argument(
        '-a', '--all-solutions', action='store_true',
        help='return all solutions (if supported by the solver)'
    )
    solver_options.add_argument(
        '-n', '--num-solutions', type=int, help='number of solutions to return'
    )
    solver_options.add_argument(
        '-t', '--timeout', type=int,
        help='time limit for flattening and solving (in seconds)'
    )
    solver_options.add_argument(
        '-p', '--parallel', type=int,
        help='the number of threads the solver should use'
    )
    solver_options.add_argument(
        '-f', '--free-search', action='store_true',
        help='instruct the solver to run a free search'
    )
    solver_options.add_argument(
        '-r', '--seed', type=int, help='the random seed for the user'
    )
    solver_options.add_argument(
        '--solver-args', type=ast.literal_eval, default={},
        help='additional arguments to pass to the solver'
    )

    output_options = minizinc_parser.add_argument_group('Output options')
    output_options.add_argument(
        '-o', '--output-file',
        help='path to the output file (default print on standard output)'
    )
    output_options.add_argument(
        '-l', '--solver-log', action='store_true', help='print solver\'s log'
    )
    output_options.add_argument(
        '--output-mode', choices=['dict', 'item', 'dzn', 'json', 'raw'],
        default='dict', help='the output format'
    )
    output_options.add_argument(
        '--output-objective', action='store_true',
        help='print the value of the objective'
    )
    output_options.add_argument(
        '--non-unique', action='store_true',
        help='allow for non unique solutions'
    )
    output_options.add_argument(
        '--allow-multiple-assignments', action='store_true',
        help='equivalent to MiniZinc allow-multiple-assigments option'
    )
    output_options.add_argument(
        '--no-declare-enums', action='store_true',
        help='do not declare enums when serializing enum values'
    )

    minizinc_parser.set_defaults(func=_minizinc)


    # config
    config_parser = subparsers.add_parser(
        'config', help='config pymzn variables'
    )
    config_parser.add_argument(
        'key', nargs='?', help='the property to get/set'
    )
    config_parser.add_argument(
        '-d', '--delete', action='store_true', help='delete a key'
    )
    config_parser.add_argument(
        'value', nargs='?', help='the value(s) to set'
    )
    config_parser.set_defaults(func=_config)

    args = parser.parse_args()

    debug(args.verbose)

    args = vars(args)
    if 'solver_args' in args:
        args.update(args['solver_args'])

    args['func'](**args)


if __name__ == '__main__':
    main()




"""\
PyMzn allows to execute minizinc concurrently within an event loop started by
Python's `asyncio` package. This allows to asyncronously solve one or more
models and collect their solutions as they are produced by the solver. This
could be useful in several situations, especially if the model returns a very
long stream of solutions or it takes more than few seconds to complete.
Executing the model asynchronously does not block the execution of the main
thread while waiting for the solver to terminate and provides a way to read
solutions as soon as they are found by the solver.

Consider the following simple model::

    %% configurations.mzn %%

    int: N = 7;
    array[1 .. N] of var 1 .. N: x;

    solve satisfy;

Suppose that you want to get all the solutions of the above problem, but you
need to do some other task while the solver does its work. We can execute
`minizinc` concurrently in this way::

    import pymzn
    import asyncio
    from pymzn.aio import minizinc

    async def run():
        solns = await minizinc('test.mzn', all_solutions=True)
        while solns.status is not pymzn.Status.COMPLETE:
            print('Performing some other task ...')
            await asyncio.sleep(1)
            for soln in solns:
                print(soln)

    if __name__ == '__main__':
        asyncio.run(run())


"""


try:
    from .mzn import aio as _aio
    from .mzn.aio import *
    __all__ = _aio.__all__
except SyntaxError as err:
    raise ImportError(
        'You need Python 3.6 or higher to use the pymzn.aio package.'
    ) from err





from time import monotonic as _time
from asyncio.subprocess import create_subprocess_exec, PIPE


__all__ = ['start_process']


class ProcessWrapper:

    def __init__(self, proc):
        self._proc = proc
        self.start_time = _time()
        self.end_time = None
        self.stdout_data = None
        self.stderr_data = None
        self._finalized = False

    def __repr__(self):
        return repr(self._proc)

    @property
    def returncode(self):
        return self._proc.returncode

    async def wait(self):
        return await self._proc.wait()

    def send_signal(self, signal):
        self._proc.send_signal(signal)

    def terminate(self):
        self._proc.terminate()

    def kill(self):
        self._proc.kill()

    @property
    def stdout(self):
        return self._stdout_stream

    @property
    def stderr(self):
        return self._stderr_stream

    async def read(self):
        if self._finalized:
            return self.stdout_data

        try:
            stdout, stderr = await self._proc.communicate()
            self.stdout_data = stdout
            self.stderr_data = stderr
        except:
            if self.returncode is None:
                try:
                    self.terminate()
                    await self.wait()
                except:
                    self.kill()
                    raise
            raise
        finally:
            self.end_time = _time()
            self._finalized = True

    async def readlines(self):
        try:
            while not self._proc.stdout.at_eof():
                yield await self._proc.stdout.readline()
            _, stderr = await self._proc.communicate()
            self.stderr_data = stderr
        except:
            try:
                self.terminate()
                await self.wait()
            except:
                self.kill()
                raise
            raise
        finally:
            self.end_time = _time()
            self._finalized = True


async def start_process(*args, stdin=PIPE):
    return ProcessWrapper(await create_subprocess_exec(
        *args, stdin=stdin, stdout=PIPE, stderr=PIPE
    ))




# -*- coding: utf-8 -*-
"""\
PyMzn can be configured with custom executable paths and other variables.
Configuration is done via the ``pymzn.config`` object. For instance:

.. code-block:: python3

    import pymzn
    pymzn.config.set('minizinc', 'path/to/minizinc')

The configurable properties used by PyMzn are the following:

 * **minizinc**: Path to the minizinc executable;
 * **solver**: Solver instance to use when calling pymzn.minizinc;
 * **solver_args**: Arguments to pass to the solver when calling pymzn.minizinc;
 * **args**: Additional arguments to pass to the template engine;
 * **include**: List of search paths to include in all minizinc calls;
 * **keep**: Overrides the keep flag of all minizinc calls;
 * **dzn_width**: The horizontal character limit for dzn files;
   This property is used to wrap long dzn statements when writing dzn files.
   This property is also used in the minizinc function as a limit to decide
   whether to write the inline data into a file.

One can also set custom properties to be used for custom solvers.

The configuration of PyMzn can be made permanent by using the ``dump`` function
of the ``config`` object:

.. code-block:: python3

    pymzn.config.dump()

This operation, as well as loading the saved configuration file, requires the
``appdirs`` and ``pyyaml`` libraries to be installed on your system.


Debug
-----

PyMzn can also be set to print debugging messages on standard output via:

.. code-block:: python3

    pymzn.debug()

This function is meant to be used in interactive sessions or in
applications that do not configure the ``logging`` library. If you configure the
``logging`` library in your application, then PyMzn will print logging messages
as well. To disable debugging messages you can then call:

.. code-block:: python3

    pymzn.debug(False)

"""

import os


class Config(dict):

    _defaults = {
        'minizinc': 'minizinc',
        'dzn_width': 70
    }

    def __init__(self, **kwargs):
        super().__init__(**{**Config._defaults, **kwargs})

        try:
            import yaml
            cfg_file = self._cfg_file()
            if cfg_file and os.path.isfile(cfg_file):
                with open(cfg_file) as f:
                    _config = yaml.load(f)
                self.update(_config)
        except ImportError:
            pass

    def __setattr__(self, key, value):
        self[key] = value

    def __dir__(self):
        return self.keys()

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setstate__(self, state):
        pass

    def _cfg_file(self):
        try:
            import appdirs
            return os.path.join(appdirs.user_config_dir(__name__), 'config.yml')
        except ImportError:
            return None

    def dump(self):
        """Writes the changes to the configuration file."""
        try:
            import yaml
            cfg_file = self._cfg_file()
            cfg_dir, __ = os.path.split(cfg_file)
            os.makedirs(cfg_dir, exist_ok=True)
            with open(cfg_file, 'w') as f:
                yaml.dump(self, f)
        except ImportError as err:
            raise RuntimeError(
                'Cannot dump the configuration settings to file. You need to '
                'install the necessary dependencies (pyyaml, appdirs).'
            ) from err


config = Config()




"""Logging configuration."""

import logging


__all__ = ['debug', 'logger']


logger = logging.getLogger(__package__)

_debug_handler = None
logger.addHandler(logging.NullHandler())


def debug(dbg=True):
    """Enables or disables debugging messages on the standard output."""
    global _debug_handler
    if dbg and _debug_handler is None:
        _debug_handler = logging.StreamHandler()
        logger.addHandler(_debug_handler)
        logger.setLevel(logging.DEBUG)
    elif not dbg and _debug_handler is not None:
        logger.removeHandler(_debug_handler)
        _debug_handler = None
        logger.setLevel(logging.WARNING)






from enum import Enum


_space_tokens = {' ', '\t', '\n', '\r', '\f', '\v'}
State = Enum('State', 'SPACE STMT STRING COMM PRE_ML_COMM POST_ML_COMM ML_COMM')
ChunkType = Enum('ChunkType', 'SPACE STMT COMM ML_COMM')

state2chunk = {
    State.SPACE: ChunkType.SPACE,
    State.STMT: ChunkType.STMT,
    State.COMM: ChunkType.COMM,
    State.ML_COMM: ChunkType.ML_COMM
}


def chunk_model(model):
    state, prev_state = State.SPACE, None
    stmts, stmt = [], []

    for ch in model:

        if state is State.SPACE:
            if ch == '%':
                state = State.COMM
            elif ch == '/':
                prev_state = state
                state = State.PRE_ML_COMM
            elif ch == '"':
                state = State.STRING
            elif ch not in _space_tokens:
                state = State.STMT
            else:
                stmt.append(ch)

            if state is not State.SPACE:
                if len(stmt) > 0:
                    stmts.append((ChunkType.SPACE, ''.join(stmt)))
                if state is not State.PRE_ML_COMM:
                    stmt = [ch]
                else:
                    stmt = []

        elif state is State.STMT:
            if ch == '%':
                if len(stmt) > 0:
                    stmts.append((ChunkType.STMT, ''.join(stmt)))
                stmt = [ch]
                state = State.COMM
            elif ch == '/':
                prev_state = state
                state = State.PRE_ML_COMM
            elif ch == '"':
                stmt.append(ch)
                state = State.STRING
            elif ch == ';':
                stmt.append(ch)
                stmts.append((ChunkType.STMT, ''.join(stmt)))
                stmt = []
                state = State.SPACE
            else:
                stmt.append(ch)

        elif state is State.STRING:
            stmt.append(ch)
            if ch == '"':
                state = State.STMT

        elif state is State.COMM:
            if ch == '\n':
                stmts.append((ChunkType.COMM, ''.join(stmt)))
                stmt = [ch]
                state = State.SPACE
            else:
                stmt.append(ch)

        elif state is State.PRE_ML_COMM:
            if ch == '*':
                if len(stmt) > 0:
                    stmts.append((state2chunk[prev_state], ''.join(stmt)))
                stmt = ['/', ch]
                state = State.ML_COMM
            else:
                stmt += ['/', ch]
                state = prev_state
            prev_state = None
        elif state is State.ML_COMM:
            stmt.append(ch)
            if ch == '*':
                state = State.POST_ML_COMM

        elif state is State.POST_ML_COMM:
            if ch == '/':
                stmt.append(ch)
                stmts.append((ChunkType.ML_COMM, ''.join(stmt)))
                stmt = []
                state = State.SPACE
            else:
                stmt += ['*', ch]

    return stmts


def merge_statements(chunks):

    stmts = []
    in_stmt = False
    stmt = []
    for chunk_type, chunk in chunks:
        if in_stmt:
            stmt.append(chunk)
            if chunk_type is ChunkType.STMT and chunk.endswith(';'):
                stmts.append((ChunkType.STMT, ''.join(stmt)))
                stmt = []
                in_stmt = False
        elif chunk_type is ChunkType.STMT:
            if chunk.endswith(';'):
                stmts.append((chunk_type, chunk))
            else:
                stmt.append(chunk)
                in_stmt = True
        else:
            stmts.append((chunk_type, chunk))

    return stmts


def rewrap_statement(s, spaces):
    lines = []
    for line in s.splitlines():
        start = 0
        c = line[start]
        while start < len(line) and start < spaces and c in _space_tokens:
            start += 1
            c = line[start]
        lines.append(line[start:])
    return '\n'.join(lines)


def count_spaces(s, first):
    if s.startswith('\n') or first:
        leading_newlines = 0
        for c in s:
            if c == '\n':
                leading_newlines += 1
            else:
                break
        prev_space = len(s) - leading_newlines
        leading_newlines = min(leading_newlines, 2)
        space = '\n' * leading_newlines
        return space, prev_space
    else:
        return s, None


def rewrap_model(model):

    chunks = chunk_model(model)
    chunks = merge_statements(chunks)

    stmts = []
    first = True
    prev_space = None
    for chunk_type, chunk in chunks:
        if chunk_type is ChunkType.STMT:
            if prev_space:
                stmt = rewrap_statement(chunk, prev_space)
                prev_space = None
            else:
                stmt = chunk
            stmts.append(stmt)
        elif chunk_type is ChunkType.SPACE:
            space, prev_space = count_spaces(chunk, first)
            if chunk.startswith('\n\n'):
                space = '\n\n'
                prev_space = len(chunk) - 2
            elif chunk.startswith('\n'):
                space = '\n'
                prev_space = len(chunk) - 1
            else:
                space = chunk
                prev_space = None
            stmts.append(space)
        else:
            stmts.append(chunk)
            prev_space = None
        first = False

    model = ''.join(stmts)
    return model




"""\
PyMzn provides a set of functions to deal with dzn encoded strings and
files. Using these functions it is possible to serialize Python objects
into dzn format and vice-versa. For instance, the ``pymzn.dict2dzn`` function
converts an assignment of variables (provided as a dictionary) into dzn
format:

.. code-block:: python3

    pymzn.dict2dzn({'a': 2, 'b': {4, 6}, 'c': {1, 2, 3}, 'd': {3: 4.5, 4: 1.3}, 'e': [[1, 2], [3, 4], [5, 6]]})

The output is a list of dzn statements, as for the previous example:

.. code-block:: python3

    ['a = 2;', 'b = {4, 6};', 'c = 1..3;', 'd = array1d(3..4, [4.5, 1.3]);', 'e = array2d(1..3, 1..2, [1, 2, 3, 4, 5, 6];']

Optionally, you can pass the path to a dzn file where to write the
statements.

.. code-block:: python3

    pymzn.dict2dzn(data, fout='path/to/dzn')

The supported types of python objects are: ``bool``, ``int``, ``float``,
``set``, ``list``, ``dict``. Lists are converted into dzn arrays with index-set
``1 .. len(list)``. Nested lists can be cast into MiniZinc multi-dimensional
array as well. Dictionaries are converted into dzn arrays with index-set equal
to the key-set of the dictionaryy, provided that the index-set is contiguous.
Nested combinations of the previous two are also allowed, provided that the
children of every node have the same index-set. The maximum depth is 6.

To evaluate back from dzn to python objects you can use the
``pymzn.dzn2dict`` function, which takes as input a dzn formatted
string or the path to a dzn file and returns the corresponding dictionary of
variable assignments. For instance, given the following dzn file:

.. code-block:: minizinc

    %% test.dzn %%

    a = 2;
    b = {4, 6};
    c = 1..3;
    d = array1d(3..4, [4.5, 1.3]);
    e = array2d(1..3, 1..2, [1, 2, 3, 4, 5, 6]);

Running the function:

.. code-block:: python3

    pymzn.dzn2dict('test.dzn')

will return:

.. code-block:: python3

    {'a': 2, 'b': {4, 6}, 'c': {1, 2, 3}, 'd': {3: 4.5, 4: 1.3}, 'e': [[1, 2], [3, 4], [5, 6]]}

which is identical to the object we serialized in the previous example.

In general, there is a one to one translation from python objects to dzn and
back, with the only exception of arrays with index-sets not based at 1. Arrays
and matrices based at 1 are translated into lists instead of dictionaries with
explicit keys. For instance:

.. code-block:: python3

    pymzn.dict2dzn({'a': {1: 2, 2: 4, 3: 6}})
    # ['a = array1d(1..3, [2, 4, 6]);']

but when translating back the array, whose index-set is based in 1, will be
translated into a list:

.. code-block:: python3

    pymzn.dzn2dict('a = array1d(1..3, [2, 4, 6]);')
    # {'a': [2, 4, 6]}

If you wish to avoid this behavior and get all arrays as dictionaries then you
can specify ``rebase_arrays=False`` as argument to the ``pymzn.dzn2dict``
function.

If, instead, you want to rebase also the arrays and matrices with different
index-sets you can use the ``pymzn.rebase_array`` function, which will discard
the index-sets in the dictionary representation of the array (matrix) and
transform it into a list (list of lists). For instance:

.. code-block:: python3

    pymzn.rebase_array({3: {2: 1, 3: 2, 4: 3}, 4: {1: 2, 2: 3}}, recursive=True)
    # [[1, 2, 3], [2, 3]]

Since version 0.18.0, PyMzn supports conversion of MiniZinc enums too:

.. code-block:: python3

    pymzn.dzn2dict('P = {A, B, C}; x = A;')
    # {'P': {'A', 'B', 'C'}, 'x': 'A'}

Without additional information, PyMzn converts enums values into strings and
enum types into sets of strings. In order to convert MiniZinc enums into Python
enums, we need to supply an additional keyword parameter ``types``, which is a
dictionary of MiniZinc types of the variables. In this case:

.. code-block:: python3

    data = pymzn.dzn2dict('P = {A, B, C}; x = A;', types={'P': 'enum', 'x': 'P'})
    print(data)

The result will be:

.. code-block:: python3

    {'x': <P.A: 1>}

where ``P.A`` is a value of an IntEnum ``P`` with three values ``P.A``, ``P.B``
and ``P.C``. By default, PyMzn will not include the enum types into the returned
dictionary of variable assignments. You can access the enum ``P`` with:

.. code-block:: python3

    P = type(data['x'])

    print(P)
    # <enum 'P'>

    print(list(P))
    # [<P.A: 1>, <P.B: 2>, <P.C: 3>]

If you want PyMzn to return the enum types along with the variable assignments,
you can do so by setting the ``return_enums`` flag to ``True``:

.. code-block:: python3

    data = pymzn.dzn2dict('P = {A, B, C}; x = A;', types={'P': 'enum', 'x': 'P'}, return_enums=True)
    print(data)
    # {'x': <P.A: 1>, 'P': <enum 'P'>}

When executing the ``pymzn.minizinc`` function, PyMzn will take care of
retrieving the types of the output variables from the model and will
automatically convert the enum values into python objects. You can also pass the
``return_enums=True`` option to ``pymzn.minizinc`` to obtain the same result as
the above example on all solutions returned.
"""

from . import marsh
from . import parse

__all__ = marsh.__all__ + parse.__all__

from .marsh import *
from .parse import *




"""\
This package provides utilities for running synchronous external processes.
"""

import os
import subprocess

from time import monotonic as _time


__all__ = ['run_process']


class CompletedProcessWrapper:

    def __init__(self, proc, start_time, end_time):
        self._proc = proc
        self.start_time = start_time
        self.end_time = end_time

    def __repr__(self):
        return repr(self._proc)

    @property
    def args(self):
        return self._proc.args

    @property
    def returncode(self):
        return self._proc.returncode

    @property
    def stdout_data(self):
        return self._proc.stdout

    @property
    def stderr_data(self):
        return self._proc.stderr

    def readlines(self):
        yield from self.stdout_data.splitlines()


def run_process(*args, input=None):
    """Run an external process.

    Parameters
    ----------
    *args : list of str
        The arguments to pass to the external process. The first argument should
        be the executable to call.
    input : str or bytes
        The input stream to supply to the extenal process.

    Return
    ------
        Object wrapping the executed process.
    """
    shell = os.name == 'nt'
    start_time = _time()
    cp = subprocess.run(
        args, input=input, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        shell=shell, bufsize=1, universal_newlines=True
    )
    end_time = _time()
    return CompletedProcessWrapper(cp, start_time, end_time)




"""\
PyMzn supports asynchronous solving through Python coroutines. The package
``pymzn.aio`` contains the coroutine version of the standard ``pymzn.minizinc``
function. This coroutine allows to execute the minizinc as an asynchronous
process and to obtain intermediate solutions while the solver is still in
execution. This is useful e.g. when the solver may take a long time to finish
and one needs to keep track of the progress or when the number of returned
solutions is very high, so the MiniZinc process does not have to keep the
solutions in memory before the caller can start consuming them.

To use the ``minizinc`` coroutine, you need to have an event loop running in the
main thread of your application. Awaiting the ``minizinc`` coroutine produces a
lazy solution stream that, when addressed or iterated over returns all the
solutions found so far by the solver. The following is a full example of how to
use the ``minizinc`` coroutine:

.. literalinclude:: ../../../../examples/asyncronous/async.mzn
  :language: minizinc
  :caption: :download:`async.mzn <../../../../examples/asyncronous/async.mzn>`
  :name: ex-async-mzn
  :linenos:

.. literalinclude:: ../../../../examples/asyncronous/async_test.py
  :language: python3
  :caption: :download:`async.mzn <../../../../examples/asyncronous/async_test.py>`
  :name: ex-async
  :linenos:

This code will execute the ``main`` coroutine, which calls ``minizinc`` to find
all solutions of the ``async.mzn`` file. The option ``keep_solutions`` is set to
``False`` to avoid saving the solutions in memory. Each time the ``solns``
object is iterated over, it will return all the solutions found until that
point. Between iterations, a period 1 second is waited to simulate other work
performed on the main thread.
"""

from . import minizinc

__all__ = minizinc.__all__

from .minizinc import *





import asyncio
from functools import partial

from ... import config, logger

from ..solvers import gecode

from ..minizinc import (
    _minizinc_preliminaries, _flattening_args, _solve_args, _cleanup
)

from .process import start_process
from .output import AsyncSolutionParser


__all__ = ['minizinc', 'solve']


async def _start_minizinc_proc(*args, input=None):
    args = [config.minizinc] + list(args)
    logger.debug('Starting minizinc with arguments: {}'.format(args))
    return await start_process(*args, stdin=input)


async def _collect(proc, queue):
    async for line in proc.readlines():
        await queue.put(line)


def _cleanup_cb(files, task):
    _cleanup(files)


async def minizinc(
    mzn, *dzn_files, args=None, data=None, include=None, stdlib_dir=None,
    globals_dir=None, declare_enums=True, allow_multiple_assignments=False,
    keep=False, output_vars=None, output_base=None, output_mode='dict',
    solver=None, timeout=None, two_pass=None, pre_passes=None,
    output_objective=False, non_unique=False, all_solutions=False,
    num_solutions=None, free_search=False, parallel=None, seed=None,
    rebase_arrays=True, keep_solutions=True, return_enums=False,
    max_queue_size=0, **kwargs
):
    """Coroutine version of the ``pymzn.minizinc`` function.

    Parameters
    ----------
    max_queue_size : int
        Maximum number of solutions in the queue between the solution parser and
        the returned solution stream. When the queue is full, the solver
        execution will halt untill an item of the queue is consumed. This option
        is useful for memory management in cases where the solution stream gets
        very large and the caller cannot consume solutions as fast as they are
        produced. Use with care, if the full solution stream is not consumed
        before the execution of the Python program ends it may result in the
        solver becoming a zombie process. Default is ``0``, meaning an infinite
        queue.
    """

    mzn_file, dzn_files, data_file, data, keep, _output_mode, types = \
        _minizinc_preliminaries(
            mzn, *dzn_files, args=args, data=data, include=include,
            stdlib_dir=stdlib_dir, globals_dir=globals_dir,
            output_vars=output_vars, keep=keep, output_base=output_base,
            output_mode=output_mode, declare_enums=declare_enums,
            allow_multiple_assignments=allow_multiple_assignments
        )

    if not solver:
        solver = config.get('solver', gecode)

    solver_args = {**kwargs, **config.get('solver_args', {})}

    proc = await solve(
        solver, mzn_file, *dzn_files, data=data, include=include,
        stdlib_dir=stdlib_dir, globals_dir=globals_dir,
        output_mode=_output_mode, timeout=timeout, two_pass=two_pass,
        pre_passes=pre_passes, output_objective=output_objective,
        non_unique=non_unique, all_solutions=all_solutions,
        num_solutions=num_solutions, free_search=free_search, parallel=parallel,
        seed=seed, allow_multiple_assignments=allow_multiple_assignments,
        **solver_args
    )

    if output_mode == 'raw':
        solns = asyncio.Queue(maxsize=max_queue_size)
        task = asyncio.create_task(_collect(proc, solns))
    else:
        parser = AsyncSolutionParser(
            solver, output_mode=output_mode, rebase_arrays=rebase_arrays,
            types=types, keep_solutions=keep_solutions,
            return_enums=return_enums, max_queue_size=max_queue_size
        )
        solns = await parser.parse(proc)
        task = parser.parse_task

    if not keep:
        task.add_done_callback(partial(_cleanup_cb, [mzn_file, data_file]))

    return solns


async def solve(
    solver, mzn, *dzn_files, data=None, include=None, stdlib_dir=None,
    globals_dir=None, allow_multiple_assignments=False, output_mode='item',
    timeout=None, two_pass=None, pre_passes=None, output_objective=False,
    non_unique=False, all_solutions=False, num_solutions=None,
    free_search=False, parallel=None, seed=None, **kwargs
):
    args = _solve_args(
        solver, timeout=timeout, two_pass=two_pass, pre_passes=pre_passes,
        output_objective=output_objective, non_unique=non_unique,
        all_solutions=all_solutions, num_solutions=num_solutions,
        free_search=free_search, parallel=parallel, seed=seed, **kwargs
    )

    args += _flattening_args(
        mzn, *dzn_files, data=data, stdlib_dir=stdlib_dir,
        globals_dir=globals_dir, output_mode=output_mode, include=include,
        allow_multiple_assignments=allow_multiple_assignments
    )

    input = mzn if args[-1] == '-' else None
    proc = await _start_minizinc_proc(*args, input=input)
    logger.debug('Solving process started.')
    return proc




# -*- coding: utf-8 -*-

import re
import os.path

from enum import IntEnum
from collections import namedtuple
from .marsh import rebase_array


__all__ = ['IntSet', 'FloatSet', 'parse_value', 'dzn2dict']


#: Representation of a contiguous integer set
IntSet = namedtuple('IntSet', ['lb', 'ub'])

#: Representation of a contiguous float set
FloatSet = namedtuple('FloatSet', ['lb', 'ub'])


# boolean pattern
_bool_p = re.compile('^(?:true|false)$')

# integer pattern
_int_p = re.compile('^[+\-]?\d+$')

# float pattern
_float_p = re.compile('^[+\-]?\d*\.\d+(?:[eE][+\-]?\d+)?$')

# ratio pattern (used in OptiMathSat)
_ratio_p = re.compile('^\s*(?P<numerator>\d+)/(?P<denominator>\d+)$')

# enum value pattern
_enum_val_p = re.compile('^\w+$')

# contiguous set
_cont_set_p = re.compile(
    '^\(?\s*([+\-]?\s*\d+(?:\.\d+)?)\s*\)?\s*\.\.\s*\(?\s*([+\-]?\s*\d+(?:\.\d+)?)\s*\)?$'
)

# set pattern
_set_p = re.compile('^(\{(?P<vals>[\w\d\s,\.+\-\(\)]*)\})$')

# contiguous integer set pattern
_cont_int_set_p = re.compile('^([+\-]?\d+)\.\.([+\-]?\d+)$')

# enum pattern
_enum_p = re.compile('^\{(?P<vals>[\w\s,]*)\}$')

# matches any of the previous
_val_p = re.compile(
    '(?:true|false|[+\-]?\d+|[+\-]?\d*\.\d+(?:[eE][+\-]?\d+)?|\w+'
    '|\(?\s*[+\-]?\d+(?:\.\d+)?\)?\.\.\(?[+\-]?\d+(?:\.\d+)?\s*\)?'
    '|\{[\w\d\s,\.+\-\(\)]*\})'
)

# multi-dimensional array pattern
_array_p = re.compile(
    '^\s*(?:array(?P<dim>\d)d\s*\(\s*'
    '(?P<indices>[\w\d\s\.+\-\(\)\{\}]+(?:\s*,\s*[\w\d\s\.+\-\(\)\{\}]+)*)\s*,'
    '\s*)?\[(?P<vals>[\w\s\.,+\-\\\/\*^|\(\)\{\}]*)\]\)?\s*$'
)

# variable pattern
_var_p = re.compile('^\s*(?P<var>[\w]+)\s*=\s*(?P<val>.+)$', re.DOTALL)

# statement pattern
_stmt_p = re.compile('\s*([^;]+?);')

# comment pattern
_comm_p = re.compile('%.+?\n')

# set type pattern
_set_type_p = re.compile('set\s+of\s+(?P<type>\w+)')

# array type pattern
_array_type_p = re.compile(
    'array\s*\[\s*(?P<indices>\w+(\s*,\s*\w+)*)\s*\]\s+of\s+(?P<type>[\w ]+)'
)


def _parse_bool(val, raise_errors=True):
    if _bool_p.match(val):
        return {'true': True, 'false': False}[val]
    if raise_errors:
        raise ValueError('Value \'{}\' is not a Boolean.'.format(val))
    return None


def _parse_int(val, raise_errors=True):
    if _int_p.match(val):
        return int(val)
    if raise_errors:
        raise ValueError('Value \'{}\' is not a integer.'.format(val))
    return None


def _parse_float(val, raise_errors=True):
    ratio_m = _ratio_p.match(val)
    if ratio_m:
        num = float(ratio_m.group('numerator'))
        den = float(ratio_m.group('denominator'))
        return num / den
    if _float_p.match(val):
        return float(val)
    if raise_errors:
        raise ValueError('Value \'{}\' is not a float.'.format(val))
    return None


def _parse_enum_val(val, var_enum=None, raise_errors=True):
    if _enum_val_p.match(val):
        if var_enum:
            return var_enum[val]
        return val
    if raise_errors:
        raise ValueError('Value \'{}\' is not an enum value.'.format(val))
    return None


def _parse_val_basic_type(val, var_type=None, enums=None, raise_errors=True):

    if not var_type:
        return _parse_val_infer_basic_type(val, raise_errors=raise_errors)

    if 'dims' in var_type:
        if raise_errors:
            raise ValueError(
                'The type \'{}\' belongs to an array.'.format(var_type)
            )
        return None

    if 'set' in var_type and var_type['set']:
        if raise_errors:
            raise ValueError(
                'The type \'{}\' belongs to an array.'.format(var_type)
            )
        return None

    if 'enum_type' in var_type:
        enum_type = var_type['enum_type']
        val_enum = None
        if enums and enum_type in enums:
            val_enum = enums[enum_type]
        return _parse_enum_val(
            val, var_enum=val_enum, raise_errors=raise_errors
        )

    if var_type['type'] == 'bool':
        return _parse_bool(val, raise_errors=raise_errors)

    if var_type['type'] == 'int':
        return _parse_int(val, raise_errors=raise_errors)

    if var_type['type'] == 'float':
        return _parse_float(val, raise_errors=raise_errors)

    if raise_errors:
        raise ValueError('Type \'{}\' not recognized.'.format(var_type))

    return None


def _parse_val_infer_basic_type(val, raise_errors=True):

    if _bool_p.match(val):
        return {'true': True, 'false': False}[val]

    if _int_p.match(val):
        return int(val)

    ratio_m = _ratio_p.match(val)
    if ratio_m:
        num = float(ratio_m.group('numerator'))
        den = float(ratio_m.group('denominator'))
        return num / den

    if _float_p.match(val):
        return float(val)

    if _enum_val_p.match(val):
        return str(val)

    if raise_errors:
        raise ValueError('Could not parse value \'{}\'.'.format(val))

    return None


def _parse_contiguous_set(val, raise_errors=True):
    cont_set_m = _cont_set_p.match(val)
    if cont_set_m:
        lb = cont_set_m.group(1)
        ub = cont_set_m.group(2)
        if _int_p.match(lb):
            lb = int(lb)
            ub = int(ub)
            return IntSet(lb=lb, ub=ub)
        if _float_p.match(lb):
            lb = float(lb)
            ub = float(ub)
            return FloatSet(lb=lb, ub=ub)
        raise ValueError('Could not parse contiguous set \'{}\'.'.format(val))
    if raise_errors:
        raise ValueError('Value \'{}\' is not a contiguous set.'.format(val))
    return None


def _parse_set(val, set_type=None, enums=None, raise_errors=True):
    set_m = _set_p.match(val)
    if set_m:
        vals = set_m.group('vals').strip()
        if vals:
            vals_type = None
            if set_type:
                vals_type = dict(set_type)
                vals_type.pop('set', None)
            return _parse_set_vals(
                vals.split(','), vals_type=vals_type, enums=enums,
                raise_errors=raise_errors
            )
        return set()
    if raise_errors:
        raise ValueError('Value \'{}\' is not a set.'.format(val))
    return None


def _parse_set_vals(vals, vals_type=None, enums=None, raise_errors=True):
    p_s = set()
    for val in vals:
        p_val = val.strip()
        p_val = _parse_val_basic_type(
            p_val, var_type=vals_type, enums=enums, raise_errors=raise_errors
        )
        p_s.add(p_val)
    return p_s


def _parse_enum_vals(vals):
    p_s = []
    for val in vals:
        p_val = val.strip()
        if _enum_val_p.match(p_val):
            p_s.append(p_val)
        else:
            raise ValueError((
                'A value of the input set is not an enum value: {}'
            ).format(repr(p_val)), p_val)
    return p_s


def _parse_val(val, var_type=None, enums=None, raise_errors=True):

    if not var_type:
        return _parse_val_infer_type(val, raise_errors=raise_errors)

    if 'dims' in var_type:
        if raise_errors:
            raise ValueError(
                'Type \'{}\' belongs to an array.'.format(var_type)
            )
        return None

    if 'set' in var_type and var_type['set']:
        if 'enum_type' not in var_type:
            p_val = _parse_contiguous_set(val, raise_errors=False)
            if p_val is not None:
                return p_val
        return _parse_set(
            val, set_type=var_type, enums=enums, raise_errors=raise_errors
        )

    return _parse_val_basic_type(
        val, var_type=var_type, enums=enums, raise_errors=raise_errors
    )


def _parse_val_infer_type(val, raise_errors=True):

    p_val = _parse_val_infer_basic_type(val, raise_errors=False)

    if p_val is not None:
        return p_val

    cont_set_m = _cont_set_p.match(val)
    if cont_set_m:
        lb = cont_set_m.group(1)
        ub = cont_set_m.group(2)
        if _int_p.match(lb):
            lb = int(lb)
            ub = int(ub)
            return IntSet(lb=lb, ub=ub)
        if _float_p.match(lb):
            lb = float(lb)
            ub = float(ub)
            return FloatSet(lb=lb, ub=ub)
        raise ValueError('Could not parse contiguous set \'{}\'.'.format(val))

    set_m = _set_p.match(val)
    if set_m:
        vals = set_m.group('vals').strip()
        if vals:
            return _parse_set_vals(vals.split(','))
        return set()

    if raise_errors:
        raise ValueError('Could not parse value \'{}\'.'.format(val))

    return None


def _parse_array(
    val, rebase_arrays=True, var_type=None, enums=None, raise_errors=True
):
    array_m = _array_p.match(val)
    if array_m:
        vals = array_m.group('vals')
        vals = _val_p.findall(vals)
        dim = array_m.group('dim')
        if dim:  # explicit dimensions
            dim = int(dim)
            indices = array_m.group('indices')
            indices = _parse_indices(indices, enums=enums)
            assert len(indices) == dim
        else:  # assuming 1d array based in 1
            indices = [range(1, len(vals) + 1)]
        if len(vals) == 0:
            p_val = []
        else:
            vals_type = None
            if var_type:
                vals_type = dict(var_type)
                vals_type.pop('dim', None)
                vals_type.pop('dims', None)
            p_val = _parse_array_vals(
                indices, vals, rebase_arrays=rebase_arrays, vals_type=vals_type,
                enums=enums, raise_errors=raise_errors
            )
        return p_val
    if raise_errors:
        raise ValueError('Could not parse array \'{}\'.'.format(val))
    return None


def _parse_array_vals(
    indices, vals, rebase_arrays=True, vals_type=None, enums=None,
    raise_errors=True
):
    # Recursive parsing of multi-dimensional arrays of the type:
    # array2d(2..4, 1..3, [1, 2, 3, 4, 5, 6, 7, 8, 9])

    idx_set = indices[0]
    if len(indices) == 1:
        arr = {i: _parse_val(
            vals.pop(0), var_type=vals_type, enums=enums,
            raise_errors=raise_errors
        ) for i in idx_set}
    else:
        arr = {i: _parse_array_vals(
            indices[1:], vals, rebase_arrays=rebase_arrays, vals_type=vals_type,
            enums=enums, raise_errors=raise_errors
        ) for i in idx_set}

    if rebase_arrays and list(idx_set)[0] == 1:
        arr = rebase_array(arr)

    return arr


def _parse_indices(st, enums=None):
    # Parse indices of multi-dimensional arrays
    ss = st.strip().split(',')
    indices = []
    for s in ss:
        s = s.strip()
        cont_int_set_m = _cont_int_set_p.match(s)
        if cont_int_set_m:
            v1 = int(cont_int_set_m.group(1))
            v2 = int(cont_int_set_m.group(2))
            indices.append(range(v1, v2 + 1))
        elif _enum_val_p.match(s):
            if not enums or s not in enums:
                raise ValueError(
                    'No definition for enum type \'{}\' was provided.'.format(s)
                )
            # indices.append(range(1, len(enums[s]) + 1))
            indices.append(list(enums[s]))
        elif s == '{}':
            indices.append([])
        else:
            raise ValueError('Index \'{}\' is not well formatted.'.format(s))
    return indices


def parse_value(val, var_type=None, enums=None, rebase_arrays=True):
    """Parses the value of a dzn statement.

    Parameters
    ----------
    val : str
        A value in dzn format.
    var_type : dict
        The dictionary of variable type as returned by the command ``minizinc
        --model-types-only``. Default is ``None``, in which case the type of the
        variable is inferred from its value. To parse an enum value as a Python
        enum type its ``var_type`` is required, otherwise the value is simply
        returned as a string.
    enums : dict of IntEnum
        A dictionary containing Python enums, with their respective names as
        keys. These enums are available to the parser to convert enum values
        into corresponding values of their respective enum types. Enum values
        can only be parsed if also the ``var_type`` of the variable is
        available.
    rebase_arrays : bool
        If the parsed value is an array and ``rebase_arrays`` is ``True``,
        return it as zero-based lists. If ``rebase_arrays`` is ``False``,
        instead, return it as a dictionary, preserving the original index-set.

    Returns
    -------
    object
        The parsed object. The type of the object depends on the dzn value.
    """

    if not var_type:
        p_val = _parse_array(
            val, rebase_arrays=rebase_arrays, enums=enums, raise_errors=False
        )
        if p_val is not None:
            return p_val
        return _parse_val(val, enums=enums)

    if 'dims' in var_type:
        return _parse_array(
            val, rebase_arrays=rebase_arrays, var_type=var_type, enums=enums
        )

    return _parse_val(val, var_type=var_type, enums=enums)


def _to_var_type(var, var_type):
    var_type = var_type.strip()
    if var_type == 'bool':
        return {'type': 'bool'}
    elif var_type == 'int':
        return {'type': 'int'}
    elif var_type == 'float':
        return {'type': 'float'}
    elif var_type == 'enum':
        return {'type': 'int', 'set': True, 'enum_type': var}
    elif var_type.startswith('set'):
        _set_type_m = _set_type_p.match(var_type)
        set_type = _set_type_m.group('type')
        var_type = {'type': 'int', 'set': True}
        if set_type in ['bool', 'int', 'float']:
            var_type['type'] = set_type
        else:
            var_type['enum_type'] = set_type
        return var_type
    elif var_type.startswith('array'):
        _array_type_m = _array_type_p.match(var_type)
        indices = _array_type_m.group('indices')
        array_type = _array_type_m.group('type')
        array_type = array_type.strip()
        dims = [s.strip() for s in indices.split(',')]
        dim = len(dims)
        var_type = {'dim': dim, 'dims': dims}
        if array_type.startswith('set'):
            _set_type_m = _set_type_p.match(array_type)
            set_type = _set_type_m.group('type')
            var_type['type'] = 'int'
            var_type['set'] = True
            if set_type in ['bool', 'int', 'float']:
                var_type['type'] = set_type
            else:
                var_type['enum_type'] = set_type
        elif array_type in ['bool', 'int', 'float']:
            var_type['type'] = array_type
        else:
            var_type['type'] = 'int'
            var_type['enum_type'] = array_type
        return var_type
    elif re.match('\w+', var_type):
        return {'type': 'int', 'enum_type': var_type}
    else:
        err = 'Type {} of variable {} not recognized.'
        raise ValueError(err.format(var_type, var))


def dzn2dict(dzn, *, rebase_arrays=True, types=None, return_enums=False):
    """Parses a dzn string or file into a dictionary of variable assignments.

    Parameters
    ----------
    dzn : str
        A dzn content string or a path to a dzn file.
    rebase_arrays : bool
        Whether to return arrays as zero-based lists or to return them as
        dictionaries, preserving the original index-sets.
    types : dict
        Dictionary of variable types. Types can either be dictionaries, as
        returned by the ``minizinc --model-types-only``, or strings containing a
        type in dzn format. If the type is a string, it can either be the name
        of an enum type or one of the following: ``bool``, ``int``, ``float``,
        ``enum``, ``set of <type>``, ``array[<index_sets>] of <type>``. The
        default value for ``var_types`` is ``None``, in which case the type of
        most dzn assignments will be inferred automatically from the value. Enum
        values can only be parsed if their respective types are available.
    return_enums : bool
        Whether to return the parsed enum types included in the dzn content.

    Returns
    -------
    dict
        A dictionary containing the variable assignments parsed from the
        input file or string.
    """
    dzn_ext = os.path.splitext(dzn)[1]
    if dzn_ext == '.dzn':
        with open(dzn) as f:
            dzn = f.read()

    var_types = None
    if types:
        var_types = {}
        for var, var_type in types.items():
            if isinstance(var_type, str):
                var_types[var] = _to_var_type(var, var_type)
            elif isinstance(var_type, dict):
                var_types[var] = var_type
            else:
                err = 'Type of variable {} must be a string or a dict.'
                raise ValueError(err.format(var))

    enum_types = None
    if var_types:
        enum_types = []
        for var, var_type in var_types.items():
            if 'enum_type' in var_type and var_type['enum_type'] == var:
                enum_types.append(var)

    var_list = []
    dzn = _comm_p.sub('\n', dzn)
    stmts = _stmt_p.findall(dzn)
    for stmt in stmts:
        var_m = _var_p.match(stmt)
        if var_m:
            var = var_m.group('var')
            val = var_m.group('val')
            var_list.append((var, val))
        else:
            raise ValueError(
                'Unsupported parsing for statement:\n{}'.format(repr(stmt))
            )

    enums = None
    if enum_types:
        enums = {}
        remaining = []
        while len(var_list) > 0:
            var, val = var_list.pop(0)
            if var in enum_types:
                enum = None
                enum_m = _enum_p.match(val)
                if enum_m:
                    vals = enum_m.group('vals').strip()
                    if vals:
                        enum_vals = _parse_enum_vals(vals.split(','))
                        enum = IntEnum(
                            var, {v: i + 1 for i, v in enumerate(enum_vals)}
                        )
                if enum is None:
                    raise ValueError(
                        'Cannot parse enum type \'{} = {}\'.'.format(var, val)
                    )
                enums[var] = enum
            else:
                remaining.append((var, val))
        var_list = remaining

    assign = {}
    for var, val in var_list:
        var_type = None
        if var_types:
            var_type = var_types.get(var, None)
        assign[var] = parse_value(
            val, var_type=var_type, enums=enums, rebase_arrays=rebase_arrays
        )

    if return_enums and enums:
        assign.update(enums)

    return assign




# -*- coding: utf-8 -*-
"""\
PyMzn provides an interface to the ``minizinc`` executable to compile a MiniZinc
model into a FlatZinc one, solve a given problem and get back the resulting
solutions directly as Python objects.

The main function that PyMzn provides is the ``pymzn.minizinc`` function, which
executes the entire workflow for solving a constranint program encoded in
MiniZinc, just like using the ``minizinc`` executable from command line. As
added benefit, the ``pymzn.minizinc`` function takes care of adding
solver-dependent parameters and converts the solutions into Python dictionaries
by default. Solving a MiniZinc problem with PyMzn is as simple as:

.. code-block:: python3

    import pymzn
    pymzn.minizinc('test.mzn')

The ``pymzn.minizinc`` function is probably the way to go for most of the
problems, but the ``pymzn.mzn2fzn`` and ``pymzn.solns2out`` functions are also
included in the library to allow for maximum flexibility. The latter two
functions are wrappers of the two homonym MiniZinc tools for, respectively,
converting a MiniZinc model into a FlatZinc one and getting custom output from
the solution stream of a solver.
"""

import os
import re
import json
import contextlib

from time import monotonic as _time
from tempfile import NamedTemporaryFile

from .. import config, dict2dzn, logger

from .rewrap import rewrap_model
from .solvers import gecode
from .process import run_process

from . import output
from .output import *
from .output import SolutionParser


__all__ = [
    'minizinc_version', 'preprocess_model', 'save_model', 'check_model',
    'check_instance', 'minizinc', 'solve', 'mzn2fzn', 'solns2out',
    'MiniZincError'
] + output.__all__


def _run_minizinc_proc(*args, input=None):
    logger.debug('Executing minizinc with arguments: {}'.format(args))
    args = [config.minizinc] + list(args)
    return run_process(*args, input=input)


def _run_minizinc(*args, input=None):
    proc = _run_minizinc_proc(*args, input=input)
    return proc.stdout_data


def minizinc_version():
    """Returns the version of the found minizinc executable."""
    vs = _run_minizinc('--version')
    m = re.findall('version ([\d\.]+)', vs)
    if not m:
        raise RuntimeError('MiniZinc executable not found.')
    return m[0]


def check_version():
    version = minizinc_version()
    logger.info('Using MiniZinc {}.'.format(version))
    major, minor, *_ = version.split('.')
    major, minor = int(major), int(minor)
    vs = major * 100 + minor
    if vs < 202:
        raise RuntimeError('PyMzn requires MiniZinc 2.2.0 or later.')


def _process_template(model, **kwargs):
    from .templates import from_string
    return from_string(model, kwargs)


def _var_types(mzn, allow_multiple_assignments=False):
    args = ['--model-types-only']
    if allow_multiple_assignments:
        args.append('--allow-multiple-assignments')

    input = None
    if mzn.endswith('.mzn'):
        args.append(mzn)
    else:
        args.append('-')
        input = mzn

    json_str = _run_minizinc(*args, input=input)
    return json.loads(json_str)['var_types']['vars']


def _model_interface(mzn, allow_multiple_assignments=False):
    args = ['--model-interface-only']
    if allow_multiple_assignments:
        args.append('--allow-multiple-assignments')

    input = None
    if mzn.endswith('.mzn'):
        args.append(mzn)
    else:
        args.append('-')
        input = mzn

    json_str = _run_minizinc(*args, input=input)
    return json.loads(json_str)


def _dzn_output_statement(output_vars, types):
    out_var = '"{0} = ", show({0}), ";\\n"'
    out_array = '"{0} = array{1}d(", {2}, ", ", show({0}), ");\\n"'
    out_list = []
    enum_types = set()
    for var in output_vars:
        if 'enum_type' in types[var]:
            enum_types.add(types[var]['enum_type'])
        if 'dim' in types[var]:
            dims = types[var]['dims']
            if len(dims) == 1:
                dim = dims[0]
                if dim != 'int':
                    enum_types.add(dim)
                    show_idx_sets = '"{}"'.format(dim)
                else:
                    show_idx_sets = 'show(index_set({}))'.format(var)
            else:
                show_idx_sets = []
                show_idx_sets_str = 'show(index_set_{}of{}({}))'
                for d in range(1, len(dims) + 1):
                    dim = dims[d - 1]
                    if dim != 'int':
                        enum_types.add(dim)
                        show_idx_sets.append(dim)
                    else:
                        show_idx_sets.append(
                            show_idx_sets_str.format(d, len(dims), var)
                        )
                show_idx_sets = ', ", ", '.join(show_idx_sets)
            out_list.append(out_array.format(var, len(dims), show_idx_sets))
        else:
            out_list.append(out_var.format(var))

    enum_list = []
    for enum_type in enum_types:
        enum_list.append(out_var.format(enum_type))

    output = ', '.join(enum_list + out_list)
    output_stmt = 'output [{}];'.format(output)
    return output_stmt


def _process_output_vars(
    model, types, output_vars=None, allow_multiple_assignments=False
):
    if output_vars is None:
        model_int = _model_interface(
            model, allow_multiple_assignments=allow_multiple_assignments
        )
        output_vars = [k for k in model_int['output']]
    output_stmt = _dzn_output_statement(output_vars, types)
    output_stmt_p = re.compile('output\s*\[(\".+?\"|[^\"]+?)+\](\s*\+\+\s*\[(\".+?\"|[^\"]+?)+\])*\s*(?:;)?', re.DOTALL)
    if output_stmt_p.search(model):
        output_stmt = output_stmt.replace('\\', '\\\\')
        return output_stmt_p.sub(output_stmt, model)
    return '\n'.join([model, output_stmt])


def preprocess_model(model, rewrap=True, **kwargs):
    """Preprocess a MiniZinc model.

    This function takes care of preprocessing the model by resolving the
    template using the arguments passed as keyword arguments to this function.
    Optionally, this function can also "rewrap" the model, deleting spaces at
    the beginning of the lines while preserving indentation.

    Parameters
    ----------
    model : str
        The minizinc model (i.e. the content of a ``.mzn`` file).
    rewrap : bool
        Whether to "rewrap" the model, i.e. to delete leading spaces, while
        preserving indentation. Default is ``True``.
    **kwargs
        Additional arguments to pass to the template engine.

    Returns
    -------
    str
        The preprocessed model.
    """

    args = {**kwargs, **config.get('args', {})}
    model = _process_template(model, **args)

    if rewrap:
        model = rewrap_model(model)
    else:
        block_comm_p = re.compile('/\*.*\*/', re.DOTALL)
        model = block_comm_p.sub('', model)
        line_comm_p = re.compile('%.*')
        model = line_comm_p.sub('', model)

    return model


def save_model(model, output_file=None, output_dir=None, output_prefix='pymzn'):
    """Save a model to file.

    Parameters
    ----------
    model : str
        The minizinc model (i.e. the content of a ``.mzn`` file).
    output_file : str
        The path to the output file. If this parameter is ``None`` (default), a
        temporary file is created with the given model in the specified output
        directory, using the specified prefix.
    output_dir : str
        The directory where to create the file in case ``output_file`` is None.
        Default is ``None``, which creates a file in the system temporary directory.
    output_prefix : str
        The prefix for the output file if created. Default is ``'pymzn'``.

    Returns
    -------
    str
        The path to the newly created ``.mzn`` file.
    """
    if output_file:
        mzn_file = output_file
        output_file = open(output_file, 'w+', buffering=1)
    else:
        output_prefix += '_'
        output_file = NamedTemporaryFile(
            dir=output_dir, prefix=output_prefix, suffix='.mzn', delete=False,
            mode='w+', buffering=1
        )
        mzn_file = output_file.name

    output_file.write(model)
    output_file.close()

    logger.info('Generated file {}'.format(mzn_file))
    return mzn_file


def _cleanup(files):
    with contextlib.suppress(FileNotFoundError):
        for _file in files:
            if _file:
                os.remove(_file)
                logger.info('Deleted file: {}'.format(_file))


def _prepare_data(mzn_file, data, keep_data=False, declare_enums=True):
    if not data:
        return None, None

    if isinstance(data, dict):
        data = dict2dzn(data, declare_enums=declare_enums)
    elif isinstance(data, str):
        data = [data]
    elif not isinstance(data, list):
        raise TypeError('The additional data provided is not valid.')

    if keep_data or sum(map(len, data)) >= int(config.dzn_width):
        mzn_base, __ = os.path.splitext(mzn_file)
        data_file = mzn_base + '_data.dzn'
        with open(data_file, 'w') as f:
            f.write('\n'.join(data))
        logger.debug('Generated file: {}'.format(data_file))
        data = None
    else:
        data = ' '.join(data)
        data_file = None
    return data, data_file


def _flattening_args(
    mzn, *dzn_files, data=None, stdlib_dir=None, globals_dir=None,
    output_mode='dict', include=None, no_ozn=False, output_base=None,
    allow_multiple_assignments=False
):
    args = []

    if stdlib_dir:
        args += ['--stdlib_dir', stdlib_dir]
    if globals_dir:
        args += ['-G', globals_dir]
    if output_mode and output_mode in ['dzn', 'json', 'item']:
        args += ['--output-mode', output_mode]
    if no_ozn:
        args.append('--no-output-ozn')
    if output_base:
        args += ['--output-base', output_base]
    if allow_multiple_assignments:
        args.append('--allow-multiple-assignments')

    if include:
        if isinstance(include, str):
            include = [include]
        elif not isinstance(include, list):
            raise TypeError('The include path is not valid.')
    else:
        include = []

    include += config.get('include', [])
    for path in include:
        args += ['-I', path]

    if data:
        args += ['-D', data]

    if mzn.endswith('.mzn'):
        args += [mzn] + list(dzn_files)
    else:
        args += list(dzn_files) + ['-']

    return args


def check_instance(
    mzn, *dzn_files, data=None, include=None, stdlib_dir=None, globals_dir=None,
    allow_multiple_assignments=False
):
    """Perform instance checking on a model + data.

    This function calls the command ``minizinc --instance-check-only`` to check
    for consistency of the given model + data.

    Parameters
    ----------
    mzn : str
        The minizinc model. This can be either the path to the ``.mzn`` file or
        the content of the model itself.
    *dzn_files
        A list of paths to dzn files to attach to the minizinc execution,
        provided as positional arguments; by default no data file is attached.
    data : dict
        Additional data as a list of strings containing dzn variables
        assignments.
    include : str or list
        One or more additional paths to search for included ``.mzn`` files.
    stdlib_dir : str
        The path to the MiniZinc standard library. Provide it only if it is
        different from the default one.
    globals_dir : str
        The path to the MiniZinc globals directory. Provide it only if it is
        different from the default one.
    allow_multiple_assignments : bool
        Whether to allow multiple assignments of variables. Sometimes is
        convenient to simply let the data file override the value already
        assigned in the minizinc file. Default is ``False``.

    Raises
    ------
        ``MiniZincError`` if instance checking fails.
    """

    args = ['--instance-check-only']
    args += _flattening_args(
        mzn, *dzn_files, data=data, include=include, stdlib_dir=stdlib_dir,
        globals_dir=globals_dir,
        allow_multiple_assignments=allow_multiple_assignments
    )

    input = mzn if args[-1] == '-' else None
    proc = _run_minizinc_proc(*args, input=input)

    if proc.stderr_data:
        raise MiniZincError(
            mzn if input is None else '\n' + mzn + '\n', args, proc.stderr_data
        )


def check_model(
    mzn, *, include=None, stdlib_dir=None, globals_dir=None
):
    """Perform model checking on a given model.

    This function calls the command ``minizinc --model-check-only`` to check
    for consistency of the given model.

    Parameters
    ----------
    mzn : str
        The minizinc model. This can be either the path to the ``.mzn`` file or
        the content of the model itself.
    include : str or list
        One or more additional paths to search for included ``.mzn`` files.
    stdlib_dir : str
        The path to the MiniZinc standard library. Provide it only if it is
        different from the default one.
    globals_dir : str
        The path to the MiniZinc globals directory. Provide it only if it is
        different from the default one.

    Raises
    ------
        ``MiniZincError`` if model checking fails.
    """

    args = ['--model-check-only']
    args += _flattening_args(
        mzn, include=include, stdlib_dir=stdlib_dir, globals_dir=globals_dir
    )

    input = mzn if args[-1] == '-' else None
    proc = _run_minizinc_proc(*args, input=input)

    if proc.stderr_data:
        raise MiniZincError(
            mzn if input is None else '\n' + mzn + '\n', args, proc.stderr_data
        )


def _minizinc_preliminaries(
    mzn, *dzn_files, args=None, data=None, include=None, stdlib_dir=None,
    globals_dir=None, output_vars=None, keep=False, output_base=None,
    output_mode='dict', declare_enums=True, allow_multiple_assignments=False
):

    check_version()

    if mzn and isinstance(mzn, str):
        if mzn.endswith('mzn'):
            if os.path.isfile(mzn):
                mzn_file = mzn
                with open(mzn) as f:
                    model = f.read()
            else:
                raise ValueError('The file does not exist.')
        else:
            mzn_file = None
            model = mzn
    else:
        raise TypeError(
            'The mzn variable must be either the path to or the '
            'content of a MiniZinc model file.'
        )

    keep = config.get('keep', keep)

    model = preprocess_model(model, rewrap=keep, **(args or {}))

    check_model(
        model, include=include, stdlib_dir=stdlib_dir, globals_dir=globals_dir
    )

    types = _var_types(
        model, allow_multiple_assignments=allow_multiple_assignments
    )

    if output_mode == 'dict':
        model = _process_output_vars(
            model, types, output_vars,
            allow_multiple_assignments=allow_multiple_assignments
        )

    output_dir = None
    output_prefix = 'pymzn'
    if keep:
        if output_base:
            output_dir, output_prefix = os.path.split(output_base)
        else:
            mzn_dir = os.getcwd()
            if mzn_file:
                mzn_dir, mzn_name = os.path.split(mzn_file)
                output_prefix, _ = os.path.splitext(mzn_name)
            output_dir = mzn_dir

    mzn_file = save_model(
        model, output_dir=output_dir, output_prefix=output_prefix
    )

    dzn_files = list(dzn_files)
    data, data_file = _prepare_data(
        mzn_file, data, keep, declare_enums=declare_enums
    )
    if data_file:
        dzn_files.append(data_file)

    check_instance(
        model, *dzn_files, data=data, include=include, stdlib_dir=stdlib_dir,
        globals_dir=globals_dir,
        allow_multiple_assignments=allow_multiple_assignments
    )

    if output_mode == 'dict':
        _output_mode = 'item'
    else:
        _output_mode = output_mode

    return mzn_file, dzn_files, data_file, data, keep, _output_mode, types


def minizinc(
    mzn, *dzn_files, args=None, data=None, include=None, stdlib_dir=None,
    globals_dir=None, declare_enums=True, allow_multiple_assignments=False,
    keep=False, output_vars=None, output_base=None, output_mode='dict',
    solver=None, timeout=None, two_pass=None, pre_passes=None,
    output_objective=False, non_unique=False, all_solutions=False,
    num_solutions=None, free_search=False, parallel=None, seed=None,
    rebase_arrays=True, keep_solutions=True, return_enums=False, **kwargs
):
    """Implements the workflow for solving a CSP problem encoded with MiniZinc.

    Parameters
    ----------
    mzn : str
        The minizinc model. This can be either the path to the ``.mzn`` file or
        the content of the model itself.
    *dzn_files
        A list of paths to dzn files to attach to the minizinc execution,
        provided as positional arguments; by default no data file is attached.
    args : dict
        Arguments for the template engine.
    data : dict
        Additional data as a dictionary of variables assignments to supply to
        the minizinc executable. The dictionary is automatically converted to
        dzn format by the ``pymzn.dict2dzn`` function.
    include : str or list
        One or more additional paths to search for included ``.mzn`` files.
    stdlib_dir : str
        The path to the MiniZinc standard library. Provide it only if it is
        different from the default one.
    globals_dir : str
        The path to the MiniZinc globals directory. Provide it only if it is
        different from the default one.
    declare_enums : bool
        Whether to declare enum types when converting inline data into dzn
        format. If the enum types are declared elsewhere this option should be
        False. Default is ``True``.
    allow_multiple_assignments : bool
        Whether to allow multiple assignments of variables. Sometimes is
        convenient to simply let the data file override the value already
        assigned in the minizinc file. Default is ``False``.
    keep : bool
        Whether to keep the generated ``.mzn``, ``.dzn``, ``.fzn`` and ``.ozn``
        files or not. If False, the generated files are created as temporary
        files which will be deleted right after the problem is solved. Though
        files generated by PyMzn are not intended to be kept, this property can
        be used for debugging purpose. Note that in case of error the files are
        not deleted even if this parameter is ``False``. Default is ``False``.
    output_vars : list of str
        A list of output variables. These variables will be the ones included in
        the output dictionary. Only available if ``ouptut_mode='dict'``.
    output_base : str
        Output directory for the files generated by PyMzn. The default
        (``None``) is the temporary directory of your OS (if ``keep=False``) or
        the current working directory (if ``keep=True``).
    output_mode : {'dict', 'item', 'dzn', 'json', 'raw'}
        The desired output format. The default is ``'dict'`` which returns a
        stream of solutions decoded as python dictionaries. The ``'item'``
        format outputs a stream of strings as returned by the ``solns2out``
        tool, formatted according to the output statement of the MiniZinc model.
        The ``'dzn'`` and ``'json'`` formats output a stream of strings
        formatted in dzn of json respectively. The ``'raw'`` format, instead
        returns the whole solution stream, without parsing.
    solver : Solver
        The ``Solver`` instance to use. The default solver is ``gecode``.
    timeout : int
        The timeout in seconds for the flattening + solving process.
    two_pass : bool or int
        If ``two_pass`` is True, then it is equivalent to the ``--two-pass``
        option for the ``minizinc`` executable. If ``two_pass`` is an integer
        ``<n>``, instead, it is equivalent to the ``-O<n>`` option for the
        ``minizinc`` executable.
    pre_passes : int
        Equivalent to the ``--pre-passes`` option for the ``minizinc``
        executable.
    output_objective : bool
        Equivalent to the ``--output-objective`` option for the ``minizinc``
        executable. Adds a field ``_objective`` to all solutions.
    non_unique : bool
        Equivalent to the ``--non-unique`` option for the ``minizinc``
        executable.
    all_solutions : bool
        Whether all the solutions must be returned. This option might not work
        if the solver does not support it. Default is ``False``.
    num_solutions : int
        The upper bound on the number of solutions to be returned. This option
        might not work if the solver does not support it. Default is ``1``.
    free_search : bool
        If ``True``, instruct the solver to perform free search.
    parallel : int
        The number of parallel threads the solver can utilize for the solving.
    seed : int
        The random number generator seed to pass to the solver.
    rebase_arrays : bool
        Whether to "rebase" parsed arrays (see the `Dzn files
        <http://paolodragone.com/pymzn/reference/dzn>`__ section). Default is
        True.
    keep_solutions : bool
        Whether to store the solutions in memory after solving is done. If
        ``keep_solutions`` is ``False``, the returned solution stream can only
        be iterated once and cannot be addressed as a list.
    return_enums : bool
        Wheter to return enum types along with the variable assignments in the
        solutions. Only used if ``output_mode='dict'``. Default is ``False``.
    **kwargs
        Additional arguments to pass to the solver, provided as additional
        keyword arguments to this function. Check the solver documentation for
        the available arguments.

    Returns
    -------
    Solutions or str
        If ``output_mode`` is not ``'raw'``, returns a list-like object
        containing the solutions found by the solver. The format of the solution
        depends on the specified ``output_mode``. If ``keep_solutions=False``,
        the returned object cannot be addressed as a list and can only be
        iterated once. If ``output_mode='raw'``, the function returns the whole
        solution stream as a single string.
    """

    mzn_file, dzn_files, data_file, data, keep, _output_mode, types = \
        _minizinc_preliminaries(
            mzn, *dzn_files, args=args, data=data, include=include,
            stdlib_dir=stdlib_dir, globals_dir=globals_dir,
            output_vars=output_vars, keep=keep, output_base=output_base,
            output_mode=output_mode, declare_enums=declare_enums,
            allow_multiple_assignments=allow_multiple_assignments
        )

    if not solver:
        solver = config.get('solver', gecode)

    solver_args = {**kwargs, **config.get('solver_args', {})}

    proc = solve(
        solver, mzn_file, *dzn_files, data=data, include=include,
        stdlib_dir=stdlib_dir, globals_dir=globals_dir,
        output_mode=_output_mode, timeout=timeout, two_pass=two_pass,
        pre_passes=pre_passes, output_objective=output_objective,
        non_unique=non_unique, all_solutions=all_solutions,
        num_solutions=num_solutions, free_search=free_search, parallel=parallel,
        seed=seed, allow_multiple_assignments=allow_multiple_assignments,
        **solver_args
    )

    if not keep:
        _cleanup([mzn_file, data_file])

    if output_mode == 'raw':
        return proc.stdout_data

    parser = SolutionParser(
        solver, output_mode=output_mode, rebase_arrays=rebase_arrays,
        types=types, keep_solutions=keep_solutions, return_enums=return_enums
    )
    solns = parser.parse(proc)
    return solns


def _solve_args(
    solver, timeout=None, two_pass=None, pre_passes=None,
    output_objective=False, non_unique=False, all_solutions=False,
    num_solutions=None, free_search=False, parallel=None, seed=None, **kwargs
):
    args = []
    if timeout:
        args += ['--time-limit', timeout * 1000] # minizinc takes milliseconds

    if two_pass:
        if isinstance(two_pass, bool):
            args.append('--two-pass')
        elif isinstance(two_pass, int):
            args.append('-O{}'.format(two_pass))

    if pre_passes:
        args += ['--pre-passes', str(pre_passes)]

    if output_objective:
        args.append('--output-objective')

    if non_unique:
        args.append('--non-unique')

    args += ['--solver', solver.solver_id]
    args += solver.args(
        all_solutions=all_solutions, num_solutions=num_solutions,
        free_search=free_search, parallel=parallel, seed=seed, **kwargs
    )

    return args


def solve(
    solver, mzn, *dzn_files, data=None, include=None, stdlib_dir=None,
    globals_dir=None, allow_multiple_assignments=False, output_mode='item',
    timeout=None, two_pass=None, pre_passes=None, output_objective=False,
    non_unique=False, all_solutions=False, num_solutions=None,
    free_search=False, parallel=None, seed=None, **kwargs
):
    """Flatten and solve a MiniZinc program.

    Parameters
    ----------
    solver : Solver
        The ``Solver`` instance to use.
    mzn : str
        The path to the minizinc model file.
    *dzn_files
        A list of paths to dzn files to attach to the minizinc execution,
        provided as positional arguments; by default no data file is attached.
    data : list of str
        Additional data as a list of strings containing dzn variables
        assignments.
    include : str or list
        One or more additional paths to search for included ``.mzn`` files.
    stdlib_dir : str
        The path to the MiniZinc standard library. Provide it only if it is
        different from the default one.
    globals_dir : str
        The path to the MiniZinc globals directory. Provide it only if it is
        different from the default one.
    allow_multiple_assignments : bool
        Whether to allow multiple assignments of variables. Sometimes is
        convenient to simply let the data file override the value already
        assigned in the minizinc file. Default is ``False``.
    output_mode : {'item', 'dzn', 'json'}
        The desired output format. The default is ``'item'`` which outputs a
        stream of strings as returned by the ``solns2out`` tool, formatted
        according to the output statement of the MiniZinc model. The ``'dzn'``
        and ``'json'`` formats output a stream of strings formatted in dzn and
        json respectively.
    timeout : int
        The timeout in seconds for the flattening + solving process.
    two_pass : bool or int
        If ``two_pass`` is True, then it is equivalent to the ``--two-pass``
        option for the ``minizinc`` executable. If ``two_pass`` is an integer
        ``<n>``, instead, it is equivalent to the ``-O<n>`` option for the
        ``minizinc`` executable.
    pre_passes : int
        Equivalent to the ``--pre-passes`` option for the ``minizinc``
        executable.
    output_objective : bool
        Equivalent to the ``--output-objective`` option for the ``minizinc``
        executable. Adds a field ``_objective`` to all solutions.
    non_unique : bool
        Equivalent to the ``--non-unique`` option for the ``minizinc``
        executable.
    all_solutions : bool
        Whether all the solutions must be returned. This option might not work
        if the solver does not support it. Default is ``False``.
    num_solutions : int
        The upper bound on the number of solutions to be returned. This option
        might not work if the solver does not support it. Default is ``1``.
    free_search : bool
        If True, instruct the solver to perform free search.
    parallel : int
        The number of parallel threads the solver can utilize for the solving.
    seed : int
        The random number generator seed to pass to the solver.
    **kwargs
        Additional arguments to pass to the solver, provided as additional
        keyword arguments to this function. Check the solver documentation for
        the available arguments.

    Returns
    -------
        Object wrapping the executed process.
    """

    args = _solve_args(
        solver, timeout=timeout, two_pass=two_pass, pre_passes=pre_passes,
        output_objective=output_objective, non_unique=non_unique,
        all_solutions=all_solutions, num_solutions=num_solutions,
        free_search=free_search, parallel=parallel, seed=seed, **kwargs
    )

    args += _flattening_args(
        mzn, *dzn_files, data=data, stdlib_dir=stdlib_dir,
        globals_dir=globals_dir, output_mode=output_mode, include=include,
        allow_multiple_assignments=allow_multiple_assignments
    )

    input = mzn if args[-1] == '-' else None

    t0 = _time()

    try:
        proc = _run_minizinc_proc(*args, input=input)
    except RuntimeError as err:
        raise MiniZincError(mzn_file, args) from err

    solve_time = _time() - t0
    logger.info('Solving completed in {:>3.2f} sec'.format(solve_time))

    return proc


def mzn2fzn(
    mzn, *dzn_files, args=None, data=None, include=None, stdlib_dir=None,
    globals_dir=None, declare_enums=True, allow_multiple_assignments=False,
    keep=False, output_vars=None, output_base=None, output_mode='item',
    no_ozn=False
):
    """Flatten a MiniZinc model into a FlatZinc one.

    This function is equivalent to the command ``minizinc --compile``.

    Parameters
    ----------
    mzn : str
        The minizinc model. This can be either the path to the ``.mzn`` file or
        the content of the model itself.
    *dzn_files
        A list of paths to dzn files to attach to the minizinc execution,
        provided as positional arguments; by default no data file is attached.
    args : dict
        Arguments for the template engine.
    data : dict
        Additional data as a dictionary of variables assignments to supply to
        the minizinc executable. The dictionary is automatically converted to
        dzn format by the ``pymzn.dict2dzn`` function.
    include : str or list
        One or more additional paths to search for included ``.mzn`` files.
    stdlib_dir : str
        The path to the MiniZinc standard library. Provide it only if it is
        different from the default one.
    globals_dir : str
        The path to the MiniZinc globals directory. Provide it only if it is
        different from the default one.
    declare_enums : bool
        Whether to declare enum types when converting inline data into dzn
        format. If the enum types are declared elsewhere this option should be
        False. Default is ``True``.
    allow_multiple_assignments : bool
        Whether to allow multiple assignments of variables. Sometimes is
        convenient to simply let the data file override the value already
        assigned in the minizinc file. Default is ``False``.
    keep : bool
        Whether to keep the generated ``.mzn``, ``.dzn``, ``.fzn`` and ``.ozn``
        files or not. If False, the generated files are created as temporary
        files which will be deleted right after the problem is solved. Though
        files generated by PyMzn are not intended to be kept, this property can
        be used for debugging purpose. Note that in case of error the files are
        not deleted even if this parameter is ``False``. Default is ``False``.
    output_vars : list of str
        A list of output variables. These variables will be the ones included in
        the output dictionary. Only available if ``ouptut_mode='dict'``.
    output_base : str
        Output directory for the files generated by PyMzn. The default
        (``None``) is the temporary directory of your OS (if ``keep=False``) or
        the current working directory (if ``keep=True``).
    output_mode : {'dict', 'item', 'dzn', 'json', 'raw'}
        The desired output format. The default is ``'dict'`` which returns a
        stream of solutions decoded as python dictionaries. The ``'item'``
        format outputs a stream of strings as returned by the ``solns2out``
        tool, formatted according to the output statement of the MiniZinc model.
        The ``'dzn'`` and ``'json'`` formats output a stream of strings
        formatted in dzn and json respectively. The ``'raw'`` format, instead
        returns the whole solution stream, without parsing.
    no_ozn : bool
        If ``True``, the ozn file is not produced, ``False`` otherwise.

    Returns
    -------
    tuple (str, str)
        The paths to the generated fzn and ozn files. If ``no_ozn=True``, the
        second argument is ``None``.
    """

    mzn_file, dzn_files, data_file, data, keep, _output_mode, types = \
        _minizinc_preliminaries(
            mzn, *dzn_files, args=args, data=data, include=include,
            stdlib_dir=stdlib_dir, globals_dir=globals_dir,
            output_vars=output_vars, keep=keep, output_base=output_base,
            output_mode=output_mode, declare_enums=declare_enums,
            allow_multiple_assignments=allow_multiple_assignments
        )

    args = ['--compile']
    args += _flattening_args(
        mzn_file, *dzn_files, data=data, stdlib_dir=stdlib_dir,
        globals_dir=globals_dir, output_mode=output_mode, include=include,
        no_ozn=no_ozn, output_base=output_base,
        allow_multiple_assignments=allow_multiple_assignments
    )

    t0 = _time()
    _run_minizinc(*args)
    flattening_time = _time() - t0
    logger.info('Flattening completed in {:>3.2f} sec'.format(flattening_time))

    if not keep:
        with contextlib.suppress(FileNotFoundError):
            if data_file:
                os.remove(data_file)
                logger.info('Deleted file: {}'.format(data_file))

    if output_base:
        mzn_base = output_base
    else:
        mzn_base = os.path.splitext(mzn_file)[0]

    fzn_file = '.'.join([mzn_base, 'fzn'])
    fzn_file = fzn_file if os.path.isfile(fzn_file) else None
    ozn_file = '.'.join([mzn_base, 'ozn'])
    ozn_file = ozn_file if os.path.isfile(ozn_file) else None

    if fzn_file:
        logger.info('Generated file: {}'.format(fzn_file))
    if ozn_file:
        logger.info('Generated file: {}'.format(ozn_file))

    return fzn_file, ozn_file


def solns2out(stream, ozn_file):
    """Wraps the ``solns2out`` utility, executes it on the solution stream, and
    then returns the output stream.

    Parameters
    ----------
    stream : str
        A solution stream. It may be a solution stream saved by a previous call
        to minizinc.
    ozn_file : str
        The path to the ``.ozn`` file produced by the ``mzn2fzn`` function.

    Returns
    -------
    str
        The output stream of solns2out encoding the solution stream according to
        the provided ozn file.
    """
    return _run_minizinc('--ozn-file', ozn_file, input=stream)


class MiniZincError(RuntimeError):
    """Generic error raised by the PyMzn functions.

    Arguments
    ---------
    mzn_file : str
        The MiniZinc file that generated the error.
    args : list of str
        The command line arguments that generated the error.
    stderr : str
        The standard error printed by the ``minizinc`` executable.
    """

    def __init__(self, mzn_file, args, stderr=None):
        self.mzn_file = mzn_file
        self.args = args
        self.stderr = stderr
        msg = (
            'An error occurred while executing minizinc on file {} '
            'with command line arguments: {}'
        ).format(mzn_file, args)
        if stderr:
            msg += '\n\n' + stderr
        super().__init__(msg)





import asyncio

from ..output import SolutionParser, Solutions


class AsyncSolutionParser(SolutionParser):

    def __init__(
        self, solver, output_mode='dict', rebase_arrays=True, types=None,
        keep_solutions=True, return_enums=False, max_queue_size=0
    ):
        super().__init__(
            solver, output_mode=output_mode, rebase_arrays=rebase_arrays,
            types=types, keep_solutions=keep_solutions,
            return_enums=return_enums
        )
        self.max_queue_size = max_queue_size
        self.parse_task = None

    async def _collect(self, solns, proc):
        async for soln in self._parse(proc):
            await solns._queue.put(soln)
        solns.status = self.status
        solns.stderr = proc.stderr_data
        solns.log = self.solver_parser.log

    async def parse(self, proc):
        solns = Solutions(
            asyncio.Queue(maxsize=self.max_queue_size), keep=self.keep_solutions
        )
        self.parse_task = asyncio.create_task(self._collect(solns, proc))
        return solns

    async def _parse(self, proc):
        parse_lines = self._parse_lines()
        parse_lines.send(None)
        async for line in proc.readlines():
            line = line.decode('utf-8')
            soln = parse_lines.send(line)
            if soln is not None:
                yield soln




import re
import os
import codecs

from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))


def find_version(*parts):
    # Open in Latin-1 so that we avoid encoding errors.
    # Use codecs.open for Python 2 compatibility
    with codecs.open(os.path.join(here, *parts), 'r', 'latin1') as f:
        version_file = f.read()

    # The version line must have the form
    # __version__ = 'ver'
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def read(*parts):
    with codecs.open(os.path.join(here, *parts), encoding='utf-8') as f:
        return f.read()


setup(
    name = 'pymzn',
    version = find_version('pymzn', '__init__.py'),
    url = 'https://github.com/paolodragone/pymzn',
    license = 'MIT',
    author = 'Paolo Dragone',
    author_email = 'dragone.paolo@gmail.com',
    description = 'A Python wrapper for the MiniZinc tool pipeline.',
    long_description = read('README.md'),
    packages = find_packages(exclude=['*tests*']),
    test_suite = "pymzn.tests",
    install_requires = [
    ],
    extra_require = {
        'file-config': ['appdirs', 'pyyaml'],
        'templates': ['jinja2']
    },
    platforms = 'any',
    classifiers = [
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    entry_points = {
        'console_scripts': [
            'pymzn=pymzn:main'
        ]
    }
)



# -*- coding: utf-8 -*-
"""\
PyMzn supports templating as a form of dynamic modelling. PyMzn allows to embed
code from the `Jinja2 <http://jinja.pocoo.org/>`_ templating language within a
MiniZinc model to make a PyMzn template file (usually distinguished with the
``.pmzn`` extension). An example:

.. literalinclude:: ../../../../examples/templates/knapsack.pmzn
  :language: pymzn
  :caption: :download:`knapsack.pmzn <../../../../examples/templates/knapsack.pmzn>`
  :name: ex-knapsack-comp
  :linenos:

.. literalinclude:: ../../../../examples/templates/knapsack.dzn
  :language: minizinc
  :caption: :download:`knapsack.dzn <../../../../examples/templates/knapsack.dzn>`
  :name: ex-knapsack-comp-dzn
  :linenos:

The above MiniZinc model encodes a 0-1 knapsack problem with optional
compatibility constraint. By default the template engine argument
``with_compatibility`` is ``None``, so the constraint is not enabled. In this
case, the model can be solved as usual by running:

.. code-block:: python3

    pymzn.minizinc('knapsack.pmzn', 'knapsack.dzn')

which returns:

.. code-block:: python3

    [{'x': {2, 3, 5}}]

If we want to use the compatibility constraint, we define a compatibility
matrix e.g. in a dzn file:

.. literalinclude:: ../../../../examples/templates/compatibility.dzn
  :language: minizinc
  :caption: :download:`compatibility.dzn <../../../../examples/templates/compatibility.dzn>`
  :name: ex-knapsack-comp-dzn-comp
  :linenos:

Now it is possible to pass ``with_compatibility`` argument to ``pymzn.minizinc``
function, along with the dzn file with the compatibility matrix:

.. code-block:: python3

    pymzn.minizinc('knapsack.pmzn', 'knapsack.dzn', 'compatibility.dzn', args={'with_compatibility': True})

which yields:

.. code-block:: python3

    [{'x': {1, 5}}]

As mentioned, PyMzn employs Jinja2 under the hood, so anything you can do with
Jinja2 is also possible in PyMzn, including variables, control structures,
template inheritance, and filters. PyMzn implements few custom filters as well:

- ``int(value, factor=100)`` : discretizes the given input or array,
  pre-multiplying by the given factor. Usage:
  ``{{ float_value_or_array | int}}`` or ``{{ float_value_or_array | int(factor=1000) }}``
- ``dzn(value)`` : transform the input into its equivalent dzn string. Usage:
  ``{{ dzn_argument | dzn }}``

To provide a custom search path to the template engine you can use the
function ``add_path``:

.. code-block:: python3

    pymzn.templates.add_path('path/to/templates/directory/')

This ensures that the template engine will look for imported tempates into the
provided path as well.
"""

from .. import val2dzn, logger

from copy import deepcopy
from collections.abc import Iterable


__all__ = ['from_string', 'add_path', 'add_package']


def discretize(value, factor=100):
    """Discretize the given value, pre-multiplying by the given factor"""
    if not isinstance(value, Iterable):
        return int(value * factor)
    int_value = list(deepcopy(value))
    for i in range(len(int_value)):
        int_value[i] = int(int_value[i] * factor)
    return int_value

try:
    from jinja2 import (
        Environment, BaseLoader, PackageLoader, FileSystemLoader,
        TemplateNotFound
    )
    _has_jinja = True
except ImportError:
    _has_jinja = False

if _has_jinja:

    class MultiLoader(BaseLoader):

        def __init__(self):
            self._loaders = []

        def add_loader(self, loader):
            self._loaders.append(loader)

        def get_source(self, environment, template):
            for loader in self._loaders:
                try:
                    return loader.get_source(environment, template)
                except TemplateNotFound:
                    pass
            raise TemplateNotFound(template)

        def list_templates(self):
            seen = set()
            templates = []
            for loader in self._loaders:
                for tmpl in loader.list_templates():
                    if tmpl not in seen:
                        templates.append(tmpl)
                        seen.add(tmpl)
            return templates


    _jload = MultiLoader()
    _jenv = Environment(trim_blocks=True, lstrip_blocks=True, loader=_jload)
    _jenv.filters['dzn'] = val2dzn
    _jenv.filters['int'] = discretize

_except_text = (
    '\nThe template engine is currently not available.\nTo use templates make '
    'sure Jinja2 is installed on your system.\nYou can install Jinja2 via pip:'
    '\n\n\tpip install Jinja2\n\nMore information at: '
    'http://jinja.pocoo.org/docs/intro/#installation'
)

def from_string(source, args=None):
    """Renders a template string"""
    if _has_jinja:
        logger.info('Precompiling model with arguments: {}'.format(args))
        return _jenv.from_string(source).render(args or {})
    if args:
        raise RuntimeError(_except_text)
    return source

def add_package(package_name, package_path='templates', encoding='utf-8'):
    """Adds the given package to the template search routine"""
    if not _has_jinja:
        raise RuntimeError(_except_text)
    _jload.add_loader(PackageLoader(package_name, package_path, encoding))


def add_path(searchpath, encoding='utf-8', followlinks=False):
    """Adds the given path to the template search routine"""
    if not _has_jinja:
        raise RuntimeError(_except_text)
    _jload.add_loader(FileSystemLoader(searchpath, encoding, followlinks))




# -*- coding: utf-8 -*-
"""\
PyMzn interfaces with solvers through the ``Solver`` base class. This class
includes the necessary infomation for PyMzn to setup the solver, and provides
two main functions to support custom solver arguments and parsing the solution
stream of the solver.

PyMzn provides a number of solver implementations out-of-the-box.  PyMzn's
default solver is ``pymzn.gecode``, an instance of ``pymzn.Gecode``.

To use a solver that is not provided by PyMzn or to extend an existing one, one
has to subclass the ``Solver`` class and implement the ``args`` method, which
returns a list of command line arguments for executing the process. The
command line arguments supported by this function have to be paired with proper
``extraFlags`` in the solver configuration file (see the `solver configuration
files page
<https://www.minizinc.org/doc-latest/en/fzn-spec.html#solver-configuration-files>`__
from the MiniZinc reference manual for additional details).

For instance, suppose you have the following configuration file for an external
solver, ``my_solver.msc``:

.. code-block:: json

    {
      "name" : "My Solver",
      "version": "1.0",
      "id": "org.myorg.my_solver",
      "executable": "fzn-mysolver"
    }

You want to support the command line argument ``--solve-twice-as-fast``. First
you need to add the flag into the solver configuration file:

.. code-block:: json

    {
      "name" : "My Solver",
      "version": "1.0",
      "id": "org.myorg.my_solver",
      "executable": "fzn-mysolver",
      "extraFlags": [
          ["--solve-twice-as-fast", "provides twofold speedup", "bool", "false"]
      ]
    }

This will make the argument available to the ``minizinc`` executable when using
the solver ``my_solver``. Next, to add this option to PyMzn, you need to
subclass the ``Solver`` class and override the ``args`` function:

.. code-block:: python3
   :linenos:

    from pymzn import Solver

    class MySolver(Solver):
        def __init__(self):
            super().__init__(solver_id='my_solver')

        def args(self, solve_twice_as_fast=False, **kwargs):
            args = super().args(**kwargs)
            if solve_twice_as_fast:
                args.append('--solve-twice-as-fast')
            return args

It is then possible to run the ``minizinc`` function with the custom solver:

.. code-block:: python3

    my_solver = MySolver()
    pymzn.minizinc('test.mzn', solver=my_solver, solve_twice_as_fast=True)

"""

import re


__all__ = [
    'Solver', 'Gecode', 'Chuffed', 'Optimathsat', 'Opturion', 'MIPSolver',
    'Gurobi', 'CBC', 'OscarCBLS', 'ORTools', 'gecode', 'chuffed', 'optimathsat',
    'opturion', 'gurobi', 'cbc', 'oscar_cbls', 'or_tools'
]


class Solver:
    """Abstract solver class.

    Parameters
    ----------
    solver_id : str
        The identifier to use when launching the minizinc command.
    """
    def __init__(self, solver_id):
        self.solver_id = solver_id

    def args(
        self, all_solutions=False, num_solutions=None, free_search=False,
        parallel=None, seed=None, **kwargs
    ):
        """Returns a list of command line arguments for the specified options.

        If the solver parser is able to parse statistics, this function should
        always add options to display statistics.

        Parameters
        ----------
        all_solutions : bool
            Whether all the solutions must be returned (default is False).
        num_solutions : int
            The maximum number of solutions to be returned (only used in
            satisfation problems).
        free_search : bool
            Whether the solver should be instructed to perform a free search.
        parallel : int
            The number of parallel threads the solver should use.
        seed : int
            The random number generator seed to pass to the solver.
        """
        args = ['-s', '-v']
        if all_solutions:
            args.append('-a')
        if num_solutions is not None:
            args += ['-n', num_solutions]
        if free_search:
            args.append('-f')
        if parallel is not None:
            args += ['-p', parallel]
        if seed is not None:
            args += ['-r', seed]
        return args

    class Parser:

        _line_comm_p = re.compile('%.*')

        def __init__(self):
            self._log = []

        @property
        def log(self):
            return '\n'.join(self._log)

        def parse_out(self):
            """Parse the output stream of the solver.

            This function is a generator that will receive in input the lines of
            the stdout of the minizinc solver via the send() function. For each
            line in input there should be a line in output.  This function
            should remove all the lines that are not part of the solution stream
            by yielding empty strings instead (which will be ignored by the
            solution parser).  This function should also process the lines
            following a non-standard dzn format, substituting them with
            equivalent standard dzn format (see e.g. Optimathsat).  Statistics
            should also be extracted, depending on the format of the solver, and
            then returned by the stats property.  Debug messages may be logged
            through the pymzn logger.
            """
            line = yield
            while True:
                if self._line_comm_p.match(line):
                    self._log.append(line)
                    line = yield ''
                else:
                    line = yield line

    def parser(self):
        """This function should return a new instance of the solver parser."""
        return Solver.Parser()


class Gecode(Solver):
    """Interface to the Gecode solver."""

    def __init__(self, solver_id='gecode'):
        super().__init__(solver_id)

    def args(self, fzn_flags=None, **kwargs):
        args = super().args(**kwargs)

        if fzn_flags:
            if isinstance(fzn_flags, str):
                fzn_flags = [fzn_flags]
            elif not isinstance(fzn_file, list):
                raise TypeError('Unrecognized type for fzn_flags argument.')
            args += ['--fzn-flags', ''.join(fzn_flags)]

        return args


class Chuffed(Solver):
    """Interface to the Chuffed solver."""

    def __init__(self, solver_id='chuffed'):
        super().__init__(solver_id)


class Optimathsat(Solver):
    """Interface to the Optimathsat solver."""

    def __init__(self, solver_id='optimathsat'):
        super().__init__(solver_id)

    def args(
        self, all_solutions=False, num_solutions=None, free_search=False,
        parallel=None, seed=None, **kwargs
    ):
        args = ['-input=fzn']
        return args

    class Parser(Solver.Parser):

        _rational_p = re.compile('(\d+)\/(\d+)')

        def parse_out(self):
            line = yield
            while True:
                if self._line_comm_p.match(line):
                    self._stats.append(line)
                    line = yield ''
                else:
                    for m in self._rational_p.finditer(line):
                        n, d = m.groups()
                        val = float(n) / float(d)
                        line = re.sub('{}/{}'.format(n, d), str(val), line)
                    line = yield line

    def parser(self):
        return Optimathsat.Parser()


class Opturion(Solver):
    """Interface to the Opturion CPX solver."""

    def __init__(self, solver_id='opturion'):
        super().__init__(solver_id)


class MIPSolver(Solver):
    """Generic interface to MIP solvers."""


class Gurobi(MIPSolver):
    """Interface to the Gurobi solver.

    Parameters
    ----------
    dll : str
        The string containing the dll of your gurobi installation.
    """

    def __init__(self, solver_id='gurobi', dll=None):
        super().__init__(solver_id)
        self.dll = dll

    def args(self, *args, **kwargs):
        args = super().args(*args, **kwargs)
        if self.dll is not None:
            args += ['--dll', self.dll]
        return args

    def args(
        self, all_solutions=False, num_solutions=None, free_search=False,
        parallel=None, seed=None, **kwargs
    ):
        args = []
        if all_solutions:
            args.append('-a')
        if num_solutions is not None:
            args += ['-n', num_solutions]
        if free_search:
            args.append('-f')
        if parallel is not None:
            args += ['-p', parallel]
        if self.dll is not None:
            args += ['--gurobi-dll', self.dll]
        return args


class CBC(MIPSolver):
    """Interface to the COIN-OR CBC solver."""

    def __init__(self, solver_id='osicbc'):
        super().__init__(solver_id)


class OscarCBLS(Solver):
    """Interface to the Oscar/CBLS solver."""

    def __init__(self, solver_id='oscar-cbls'):
        super().__init__(solver_id)


class ORTools(Solver):
    """Interface to the OR-tools solver.

    Parameters
    ----------
    path : str
        The path to the fzn-or-tools executable.
    globals_dir : str
        The path to the directory for global included files. You should either
        copy or link the 'share/minizinc_cp' folder from the or-tools
        distribution into the minizinc library directory (with name 'or-tools')
        or provide the full path here.
    """

    def __init__(self, solver_id='or-tools'):
        super().__init__(solver_id)


#: Default Gecode instance.
gecode = Gecode()

#: Default Chuffed instance.
chuffed = Chuffed()

#: Default Optimathsat instance.
optimathsat = Optimathsat()

#: Default Opturion instance.
opturion = Opturion()

#: Default Gurobi instance.
gurobi = Gurobi()

#: Default CBC instance.
cbc = CBC()

#: Default Oscar/CBLS instance.
oscar_cbls = OscarCBLS()

#: Default ORTools instance.
or_tools = ORTools()

